#include <fstream>
#include <iomanip>
#include <cstdlib>
#include "pthread.h"
#include "camera.h"
#include "util.h"
#define _USE_MATH_DEFINES
#include <math.h>

using namespace std;
using namespace Image;

Ccamera::Ccamera(void) {
  m_axesScale = 1.0f;
  m_maxLevel = 1;
  m_refractEpsilon = 1e-9;
}

Ccamera::~Ccamera() {
}

void Ccamera::init(const std::string cname, const int maxLevel) {
  m_cname = cname;
  m_maxLevel = maxLevel;

  // initialize camera
  m_intrinsics.resize(6);
  m_extrinsics.resize(6);

  ifstream ifstr;
  ifstr.open(cname.c_str());

  string header;
  ifstr >> header;

  cerr << "[Camera::init] Header: " << header << endl;

  if (header == "CONTOUR")
    m_txtType = 0;
  else if (header == "CONTOUR2")
    m_txtType = 2;
  else if (header == "CONTOUR3")
    m_txtType = 3;
  else if (header == "CONTOUR4")
	m_txtType = 4;
  else {
    cerr << "[Camera::init] Unrecognizable txt format" << endl;
    exit (1);
  }

  cerr << "[Camera::init] Reading vari-data" << endl;

  for (int i = 0; i < 6; ++i)
    ifstr >> m_intrinsics[i];
  for (int i = 0; i < 6; ++i)
    ifstr >> m_extrinsics[i];

  if (m_txtType == 4) {
	cerr << "Reading intrinsics" << endl;
	
	m_K.resize(maxLevel);
	m_iK.resize(maxLevel);

	for (int i = 0; i < 3; i++) {
	  for (int j = 0; j < 3; j++) {
		ifstr >> m_K[0][i][j];
	  }
	}
	invert(m_iK[0], m_K[0]);

	cerr << "Reading extrinsics" << endl;

	for (int i = 0; i < 3; i++) {
	  for (int j = 0; j < 3; j++) {
		ifstr >> m_R[i][j];
	  }
	}
	invert(m_iR, m_R);

	for (int i = 0; i < 3; i++) {
	  ifstr >> m_t[i];
	}

	cerr << "Reading refraction data" << endl;

	ifstr >> m_n[0] >> m_n[1] >> m_n[2];

	cerr << "n = " << m_n[0] << " " << m_n[1] << " " << m_n[2] << endl;

	ifstr >> m_numRefractions;

	cerr << "num_refractions = " << m_numRefractions << endl;

	m_mu.resize(m_numRefractions+1);
	m_d.resize(m_numRefractions);

	m_sumD = 0.0f;
	for (int i = 0; i < m_numRefractions; i++) {
	  ifstr >> m_mu[i] >> m_d[i];
	  m_sumD += m_d[i];
	  cerr << "mu[" << i << "]=" << m_mu[i] << " d[" << i << "]=" << m_d[i] << endl;
	}
	if (m_numRefractions > 0) {
	  ifstr >> m_mu[m_numRefractions];
	  cerr << "mu[" << m_numRefractions << "]=" << m_mu[m_numRefractions] << endl;
	}
  }
  else {
	m_numRefractions = 0;
  }

  cerr << "[Camera::init] Done reading." << endl;

  ifstr.close();
  
  //----------------------------------------------------------------------
  m_projection.resize(maxLevel);
  for (int level = 0; level < maxLevel; ++level)
    m_projection[level].resize(3);

  updateCamera();

  cerr << "[Camera::init] Done." << endl;
}

void Ccamera::updateProjection(void) {
  // Set bottom level
  setProjection(m_intrinsics, m_extrinsics, m_projection[0], m_txtType);

  for (int level = 1; level < m_maxLevel; ++level) {
    for (int i = 0; i < 3; ++i)
      m_projection[level][i] = m_projection[level - 1][i];

    m_projection[level][0] /= 2.0;
	m_projection[level][1] /= 2.0;

	if (m_txtType == 4) {
	  for (int i = 0; i < 3; ++i)
		m_K[level] = m_K[level - 1];

	  m_K[level][0] /= 2.0;
	  m_K[level][1] /= 2.0;

	  invert(m_iK[level], m_K[level]);
	}
  }
}

void Ccamera::write(const std::string file) {
  ofstream ofstr;
  ofstr.open(file.c_str());
  if (m_txtType == 0) {
    ofstr << "CONTOUR" << endl
	  << m_intrinsics[0] << ' ' << m_intrinsics[1] << ' '
	  << m_intrinsics[2] << ' ' << m_intrinsics[3] << endl
	  << m_intrinsics[4] << ' ' << m_intrinsics[5] << ' '
	  << m_extrinsics[0] << ' ' << m_extrinsics[1] << endl
	  << m_extrinsics[2] << ' ' << m_extrinsics[3] << ' '
	  << m_extrinsics[4] << ' ' << m_extrinsics[5] << endl;
  }
  else if (m_txtType == 2) {
    ofstr << "CONTOUR2" << endl;
    for (int i = 0; i < 6; ++i)
      ofstr << m_intrinsics[i] << ' ';
    ofstr << endl;
    for (int i = 0; i < 6; ++i)
      ofstr << m_extrinsics[i] << ' ';
    ofstr << endl;
  }
  else if (m_txtType == 3) {
    ofstr << "CONTOUR3" << endl;
    for (int i = 0; i < 6; ++i)
      ofstr << m_intrinsics[i] << ' ';
    ofstr << endl;
    for (int i = 0; i < 6; ++i)
      ofstr << m_extrinsics[i] << ' ';
    ofstr << endl;
  }
  else if (m_txtType == 4) {
	ofstr << "CONTOUR4" << endl
		  << m_intrinsics[0] << ' ' << m_intrinsics[1] << ' '
		  << m_intrinsics[2] << ' ' << m_intrinsics[3] << endl
		  << m_intrinsics[4] << ' ' << m_intrinsics[5] << ' '
		  << m_extrinsics[0] << ' ' << m_extrinsics[1] << endl
		  << m_extrinsics[2] << ' ' << m_extrinsics[3] << ' '
		  << m_extrinsics[4] << ' ' << m_extrinsics[5] << endl;
	// Output refraction parameters!
	ofstr << m_n[0] << ' ' << m_n[1] << ' ' << m_n[2] << endl;

	ofstr << m_numRefractions;
	for (int i = 0; i < m_numRefractions; i++) {
	  ofstr << ' ' << m_mu[i] << ' ' << m_d[i];
	}
	if (m_numRefractions > 0) {
	  ofstr << ' ' << m_mu[m_numRefractions];
	}
	ofstr << endl;
  }
  else {
    cerr << "No way. Unrecognizable format: " << m_txtType << endl;
    exit (1);
  }
  
  ofstr.close();
}

void Ccamera::updateCamera(void) {
  updateProjection();
  
  //----------------------------------------------------------------------
  m_oaxis = m_projection[0][2];
  m_oaxis[3] = 0.0;
  const float ftmp = norm(m_oaxis);
  m_oaxis[3] = m_projection[0][2][3];
  m_oaxis /= ftmp;

  m_center = getOpticalCenter();

  m_zaxis = Vec3f(m_oaxis[0], m_oaxis[1], m_oaxis[2]);
  m_xaxis = Vec3f(m_projection[0][0][0],
		  m_projection[0][0][1],
		  m_projection[0][0][2]);
  m_yaxis = cross(m_zaxis, m_xaxis);
  unitize(m_yaxis);
  m_xaxis = cross(m_yaxis, m_zaxis);
  
  Vec4f xaxis = m_projection[0][0];  xaxis[3] = 0.0f;    
  Vec4f yaxis = m_projection[0][1];  yaxis[3] = 0.0f;
  float ftmp2 = (norm(xaxis) + norm(yaxis)) / 2.0f;
  if (ftmp2 == 0.0f)
    ftmp2 = 1.0f;
  m_ipscale = ftmp2;
}

Vec4f Ccamera::getOpticalCenter(void) const {
  // orthographic case
  Vec4f ans;
  if (m_projection[0][2][0] == 0.0 && m_projection[0][2][1] == 0.0 &&
      m_projection[0][2][2] == 0.0) {
    Vec3f vtmp[2];
    for (int i = 0; i < 2; ++i)
      for (int y = 0; y < 3; ++y)
	vtmp[i][y] = m_projection[0][i][y];
	
    Vec3f vtmp2 = cross(vtmp[0], vtmp[1]);
    unitize(vtmp2);
    for (int y = 0; y < 3; ++y)
      ans[y] = vtmp2[y];
    ans[3] = 0.0;
  }
  else {
    Mat3 A;
    Vec3 b;
    for (int y = 0; y < 3; ++y) {
      for (int x = 0; x < 3; ++x)
		A[y][x] = m_projection[0][y][x];
      b[y] = - m_projection[0][y][3];
    }
    Mat3 iA;
    invert(iA, A);
    b = iA * b;
    
    for (int y = 0; y < 3; ++y)
      ans[y] = b[y];
    ans[3] = 1.0;
  }
  return ans;
}

// get scale
float Ccamera::getScale(const Vec4f& coord, const int level) const {
  if (m_maxLevel <= level) {
    cerr << "Level is not within a range: " << level << ' ' << m_maxLevel << endl;
    exit (1);
  }

  // For orthographic case
  if (m_projection[0][2][0] == 0.0 && m_projection[0][2][1] == 0.0 &&
      m_projection[0][2][2] == 0.0) {
    const Vec3f xaxis(m_projection[0][0][0], m_projection[0][0][1],
		      m_projection[0][0][2]);
    const Vec3f yaxis(m_projection[0][1][0], m_projection[0][1][1],
		      m_projection[0][1][2]);
    return (0x0001 << level) / ((xaxis.norm() + yaxis.norm()) / 2.0);
  }
  else {      
    //const float fz = coord * m_projection[level][2];    
    //return fz * (0x0001 << level) / m_ipscale;
    // ???? new by take into angle difference
    Vec4f ray = coord - m_center;
    return norm(ray) * (0x0001 << level) / m_ipscale;
  }
}

float Ccamera::getFocalX(const int level) const {
  if (m_txtType != 4) {
    cerr << "getFocalX not supported for txtType: " << m_txtType << endl;
    exit (1);
  }

  return m_K[level][0][0];
}

float Ccamera::getFocalY(const int level) const {
  if (m_txtType != 4) {
    cerr << "getFocalY not supported for txtType: " << m_txtType << endl;
    exit (1);
  }

  return m_K[level][1][1];
}

void Ccamera::setK(Mat3f& K) const{
  if (m_txtType != 2) {
    cerr << "getK not supported for txtType: " << m_txtType << endl;
    exit (1);
  }

  for (int y = 0; y < 3; ++y)
    for (int x = 0; x < 3; ++x)
      K[y][x] = 0.0;

  K[0][0] = m_intrinsics[0];
  K[1][1] = m_intrinsics[1];
  K[0][1] = m_intrinsics[2];
  K[0][2] = m_intrinsics[3];
  K[1][2] = m_intrinsics[4];
  K[2][2] = 1.0;
}

void Ccamera::setRT(Mat4f& RT) const{
  if (m_txtType != 2) {
    cerr << "getRT not supported for txtType: " << m_txtType << endl;
    exit (1);
  }

  double params[6];
  for (int i = 0; i < 6; ++i)
    params[i] = m_extrinsics[i];
    
  Mat4 RTd;
  q2proj(params, RTd);

  for (int y = 0; y < 4; ++y)
    for (int x = 0; x < 4; ++x)
      RT[y][x] = RTd[y][x];
}

void Ccamera::getR(Mat3f& R) const {
  if (m_txtType != 2) {
    cerr << "Not supported: " << m_txtType << endl;
    exit (1);
  }

  double params[6];
  for (int i = 0; i < 6; ++i)
    params[i] = m_extrinsics[i];

  Mat4 mtmp;
  q2proj(params, mtmp);
  for (int y = 0; y < 3; ++y)
    for (int x = 0; x < 3; ++x)
      R[y][x] = mtmp[y][x];
}

void Ccamera::setProjection(const std::vector<float>& intrinsics,
			   const std::vector<float>& extrinsics,
			   std::vector<Vec4f>& projection,
			   const int txtType) {
  projection.resize(3);
  double params[12];
  for (int i = 0; i < 6; ++i) {
    params[i] = intrinsics[i];
    params[6 + i] = extrinsics[i];
  }
  
  if ((txtType == 0) || (txtType == 4)) {
    for (int y = 0; y < 3; ++y) {
      for (int x = 0; x < 4; ++x ) {
		projection[y][x] = params[4 * y + x];
      }
    }
    //projection[1] = - projection[1];
  }
  else if (txtType == 2) {
    Mat4 K;
    for (int y = 0; y < 4; ++y)
      for (int x = 0; x < 4; ++x)
		K[y][x] = 0.0;

    K[0][0] = params[0];    K[1][1] = params[1];
    K[0][1] = params[2];    K[0][2] = params[3];
    K[1][2] = params[4];    K[2][2] = 1.0;
    K[3][3] = 1.0;
    
    Mat4 mtmp;
    q2proj(&params[6], mtmp);
    mtmp = K * mtmp;

    for (int y = 0; y < 3; ++y)
      for (int x = 0; x < 4; ++x)
	projection[y][x] = mtmp[y][x];
  }
  else if (txtType == 3) {
    // parameters
    // # first intrinsics
    // fovx width height 0 0 0
    // # second extrinsics
    // tx ty tz rx ry rz
    double params2[9] = {params[0], params[1], params[2],
			 params[6], params[7], params[8],
			 params[9], params[10], params[11]};
    
    setProjectionSub(params2, projection, 0);

    /*
    cout << endl;
    for (int y = 0; y < 3; ++y) {
      if (y == 1)
	cout << -projection[y] << endl;
      else
	cout << projection[y] << endl;
    }
    cout << endl;
    */
  }
  else {
    cerr << "Impossible setProjection" << endl;
    exit (1);
  }
}

void Ccamera::setProjectionSub(double params[], std::vector<Vec4f>& projection, const int level) {
  const double rx = params[6] * M_PI / 180.0;
  const double ry = params[7] * M_PI / 180.0;
  const double rz = params[8] * M_PI / 180.0;
  
  const double fovx = params[0] * M_PI / 180.0;
  
  const double f = params[1] / 2.0 / tan(fovx / 2.0);
  Mat3 K;
  K[0] = Vec3(f, 0.0, 0.0);
  K[1] = Vec3(0.0, f, 0.0);
  K[2] = Vec3(0.0, 0.0, -1.0);
  
  Mat3 trans;
  trans[0] = Vec3(1.0, 0.0, params[1] / 2.0);
  trans[1] = Vec3(0.0, -1.0, params[2] / 2.0);
  trans[2] = Vec3(0.0, 0.0, 1.0);
  
  K = trans * K;

  Mat3 Rx;
  Rx[0] = Vec3(1.0, 0.0, 0.0);
  Rx[1] = Vec3(0.0f, cos(rx), -sin(rx));
  Rx[2] = Vec3(0.0, sin(rx), cos(rx));

  Mat3 Ry;
  Ry[0] = Vec3(cos(ry), 0, sin(ry));
  Ry[1] = Vec3(0.0, 1.0, 0.0);
  Ry[2] = Vec3(-sin(ry), 0, cos(ry));

  Mat3 Rz;
  Rz[0] = Vec3(cos(rz), -sin(rz), 0.0);
  Rz[1] = Vec3(sin(rz), cos(rz), 0.0);
  Rz[2] = Vec3(0.0, 0.0, 1.0);

  //????????
  //Mat3 R = transpose(Rz) * transpose(Rx) * transpose(Ry);
  Mat3 R = transpose(Rx) * transpose(Ry) * transpose(Rz);
  
  Vec3 t(params[3], params[4], params[5]);

  Mat3 left = K * R;
  Vec3 right = - K * (R * t);

  for (int y = 0; y < 3; ++y) {
    for (int x = 0; x < 3; ++x)
      projection[y][x] = left[y][x];
    projection[y][3] = right[y];
  }

  const int scale = 0x0001 << level;
  projection[0] /= scale;
  projection[1] /= scale;
}

void Ccamera::proj2q(Mat4& mat, double q[6]) {
  double s;
  int i;

  q[3] = mat[0][3];
  q[4] = mat[1][3];
  q[5] = mat[2][3];
  q[0] = 0;
  q[1] = 0;
  q[2] = 0;
  if (mat[2][0] == 1.0) {
    q[1] = (double) -M_PI/2.0;
    q[2] = 0;
    q[0]=atan2(-mat[0][1],mat[1][1]);
  }
  else {
    if (mat[2][0] == -1.0) { 
      q[1] = M_PI/2.0;
      q[2] = 0;
      q[0]=atan2(mat[0][1],mat[1][1]);    
    }
    else {
      q[1] = (double)  asin(-mat[2][0]);
      if (cos(q[1]) > 0.0) { s = 1.0;} else { s =-1.0;};
      q[0] =atan2(mat[2][1]*s, mat[2][2]*s); 
      q[2] =atan2(mat[1][0]*s, mat[0][0]*s); 
    }
  }
  q[0]=q[0]*180/M_PI;//RadInDeg;
  q[1]=q[1]*180/M_PI;//RadInDeg;
  q[2]=q[2]*180/M_PI;//RadInDeg;
  for(i=0;i<3;i++){
    if (fabs(q[i])>180.0){
      q[i]= (q[i]>0) ? q[i]-360.0 : q[i]+360.0;
    }
  }
}

void Ccamera::q2proj(const double q[6], Mat4& mat) {
  const double a = q[0] * M_PI / 180.0;
  const double b = q[1] * M_PI / 180.0;
  const double g = q[2] * M_PI / 180.0;
  
  const double s1=sin(a);  const double s2=sin(b);  const double s3=sin(g);
  const double c1=cos(a);  const double c2=cos(b);  const double c3=cos(g);
             		
  /*   Premiere colonne*/	/*   Seconde colonne	*/
  mat[0][0]=c2*c3; 		mat[0][1]=c3*s2*s1-s3*c1;  
  mat[1][0]=s3*c2; 		mat[1][1]=s3*s2*s1+c3*c1; 
  mat[2][0]=-s2;   		mat[2][1]=c2*s1;

  /*   Troisieme colonne*/	/*  Quatrieme colonne	*/
  mat[0][2]=c3*s2*c1+s3*s1; 	mat[0][3]=q[3]; 
  mat[1][2]=s3*s2*c1-c3*s1; 	mat[1][3]=q[4]; 
  mat[2][2]=c2*c1; 		mat[2][3]=q[5];

  mat[3][0] = mat[3][1] = mat[3][2] = 0.0;
  mat[3][3] = 1.0;
}

float Ccamera::computeDepthDif(const Vec4f& lhs, const Vec4f& rhs) const {
  // orthographic projection case
  if (m_projection[0][2][0] == 0.0 && m_projection[0][2][1] == 0.0 &&
      m_projection[0][2][2] == 0.0) {
    return - m_center * (lhs - rhs);
  }
  else {
    return m_oaxis * (lhs - rhs);
  }
}

float Ccamera::computeDistance(const Vec4f& point) const {
  const float fx = point[0] - m_center[0];
  const float fy = point[1] - m_center[1];
  const float fz = point[2] - m_center[2];
  
  return sqrt(fx * fx + fy * fy + fz * fz);
}

float Ccamera::computeDepth(const Vec4f& point) const {
  // orthographic projection case
  if (m_projection[0][2][0] == 0.0 && m_projection[0][2][1] == 0.0 &&
      m_projection[0][2][2] == 0.0) {

    //cerr << "Because I'm using negative depth to represent flip. this could be a problem" << endl;
    return - m_center * point;
  }
  else {
    return m_oaxis * point;
  }
}

void Ccamera::getPAxes(const Vec4f& coord, const Vec4f& normal,
		       Vec4f& pxaxis, Vec4f& pyaxis, const int level) const {
  // yasu changed here for fpmvs
  const float pscale = getScale(coord, level);

  Vec3f normal3(normal[0], normal[1], normal[2]);
  Vec3f yaxis3 = cross(normal3, m_xaxis);
  unitize(yaxis3);
  Vec3f xaxis3 = cross(yaxis3, normal3);
  pxaxis[0] = xaxis3[0];  pxaxis[1] = xaxis3[1];  pxaxis[2] = xaxis3[2];  pxaxis[3] = 0.0;
  pyaxis[0] = yaxis3[0];  pyaxis[1] = yaxis3[1];  pyaxis[2] = yaxis3[2];  pyaxis[3] = 0.0;

  pxaxis *= pscale;
  pyaxis *= pscale;
  const float xdis = norm(project(coord + pxaxis, level) -
                          project(coord, level));
  const float ydis = norm(project(coord + pyaxis, level) -
                          project(coord, level));
  pxaxis *= m_axesScale / xdis;
  pyaxis *= m_axesScale / ydis;
  
  /*
  Vec3f normal3(normal[0], normal[1], normal[2]);

  Vec3f yaxis3 = cross(normal3, m_xaxis);
  unitize(yaxis3);
  Vec3f xaxis3 = cross(yaxis3, normal3);

  pxaxis[0] = xaxis3[0];  pxaxis[1] = xaxis3[1];
  pxaxis[2] = xaxis3[2];  pxaxis[3] = 0.0;
  pyaxis[0] = yaxis3[0];  pyaxis[1] = yaxis3[1];
  pyaxis[2] = yaxis3[2];  pyaxis[3] = 0.0;

  // ??? in order to get less noisy results with orientation optimization
  const int lessnoisy = 0;
  if (lessnoisy) {
    float xtmp = m_oaxis * pxaxis;
    float ytmp = m_oaxis * pyaxis;
    xtmp = sqrt(max(0.0f, 1.0f - xtmp * xtmp));
    ytmp = sqrt(max(0.0f, 1.0f - ytmp * ytmp));
    if (xtmp == 0.0f)
      xtmp = 0.0000001f;
    if (ytmp == 0.0f)
      ytmp = 0.0000001f;
    
    //xtmp = min(2.0f, 1.0f / xtmp);
    //ytmp = min(2.0f, 1.0f / ytmp);
    xtmp = 1.0f / xtmp;
    ytmp = 1.0f / ytmp;

    pxaxis *= xtmp;
    pyaxis *= ytmp;
  }

  // How many pixels per step
  const float scale = getScale(coord, 0) * m_axesScale;

  pxaxis *= scale;
  pyaxis *= scale;
  */
}

void Ccamera::setAxesScale(const float axesScale) {
  m_axesScale = axesScale;
}  

//void Ccamera::intersect(const Vec4f& coord, const Vec4f& abcd,
//                        Vec4f& cross, float& distance) const {
//  Vec4f ray = coord - m_center;
//  unitize(ray);
//  const float A = coord * abcd;
//  const float B = ray * abcd;
//
//  if (B == 0.0f) {
//    distance = 0xffff;
//    cross = Vec4f(0.0f, 0.0f, 0.0f, -1.0f);
//  }
//  else {
//    distance = - A / B;
//    cross = coord + distance * ray;
//  }  
//}
//
//Vec4f Ccamera::intersect(const Vec4f& coord, const Vec4f& abcd) const {
//  Vec4f ray = m_center - coord;
//
//  const float A = coord * abcd;
//  const float B = ray * abcd;
//
//  if (B == 0.0f)
//    return Vec4f(0.0f, 0.0f, 0.0f, -1.0f);
//  else
//    return coord - A / B * ray;
//}

pthread_mutex_t printMutex = PTHREAD_MUTEX_INITIALIZER;

Vec3f Ccamera::project(const Vec4f& coord, const int level, Vec4f *refrax, 
  const Vec4f *refraxInit) const {
 // if (coord != coord) {// Test for NaN
	//pthread_mutex_lock(&printMutex);
	//cerr << "[Ccamera::project] Caught a NaN coord!" << endl;
	//pthread_mutex_unlock(&printMutex);
	//exit(1);
 // }

  Vec4 q(coord);
  Vec3 vtmp; 

#if 1
  if ((m_numRefractions > 0) && 
	  (coord == coord /* avoid NaN here, rest of func will return NaN */)) {
	// Calculate the point on the nearest refractive interface to
	// the camera, then use perspective projection.

	if (q[3] != 1.0f) {
	  q /= q[3];
	}

	Vec3 ray;

	ray[0] = q[0] - m_t[0];
	ray[1] = q[1] - m_t[1];
	ray[2] = q[2] - m_t[2];

	//pthread_mutex_lock(&printMutex);
	//cerr << "[Ccamera::project] Refr coord: " << coord
	//	 << ", q: " << q 
	//	 << ", ray: " << ray
	//	 << endl;
	//pthread_mutex_unlock(&printMutex);

	// "Depth" of point along normal (perpendicular direction) of
	// the refractive interfaces
	double depth = -m_n * ray;

	if (depth >= m_sumD) {
	  // Displacement in a direction perpendicular to the normal
	  Vec3 yDir = ray + m_n * depth;
	  double yObj = norm(yDir);

	  // Result y, at a depth of 1
	  double y;

	  if (yObj == 0.0f) {
		// Ray is perpendicular to interfaces => no refraction
		y = 0.0f;
	  }
	  else {
		yDir /= yObj; // unitize it

		double lb, ub;
		double y0;

		// Compute initial solution bounds.
		if (refraxInit == NULL) {
		  // Shoot a ray directly toward the object point
		  y0 = yObj / depth;
		}
		else {
		  // Use initial guess
		  Vec3 init;

		  if ((*refraxInit)[3] != 1.0f) {
			init[0] = (*refraxInit)[0] / (*refraxInit)[3];
			init[1] = (*refraxInit)[1] / (*refraxInit)[3];
			init[2] = (*refraxInit)[2] / (*refraxInit)[3];
		  }
		  else {
			init[0] = (*refraxInit)[0];
			init[1] = (*refraxInit)[1];
			init[2] = (*refraxInit)[2];
		  }

		  y0 = (init - m_t) * yDir / m_d[0];
		}

		double y0Obj = unproject2D(depth, y0);

		//printf("num_refrac=%i sx=%0.9f sy=%0.9f ", num_refrac, sx, sy);
		//int i;
		//for (i = 0; i < num_refrac; i++) {
		//	printf("d%i=%0.9f mu%i=%0.9f\n", i, d[i], i, mu[i]);
		//}
	
		// Check the discrepancy between the actual object y displacement
		// and the back-projected displacement. Shift the intersection
		// point at the first refractive interface by the same amount, 
		// and use this as the other bound.
		if (y0Obj < yObj)
		{
			lb = y0;
			ub = y0 + (yObj - y0Obj) / m_d[0];
		}
		else {
			ub = y0;
			lb = y0 - (y0Obj - yObj) / m_d[0];
		}

		if ((ub - lb) > m_refractEpsilon) {
		  // Run optimization to find the ray direction that back-projects
		  // to the object point
		  RefProjectErrFun errFun(this, depth, yObj);
		  //pthread_mutex_lock(&printMutex);
		  //cerr << "Begin fzero call. depth = " << depth << ", coord = " << coord << "...";
		  int optRes = fzero(errFun, lb, ub, m_refractEpsilon, &y);
		  //cerr << "done!" << endl;
		  //pthread_mutex_unlock(&printMutex);

		  //pthread_mutex_lock(&printMutex);
		  //cerr << "[Ccamera::project] "
		  //	 << "lb: " << lb << " ub: " << ub << " y: " << y 
		  //	 << " yObj: " << yObj << " y0: " << y0 
		  //	 << " y0Obj: " << y0Obj << " depth: " << depth 
		  //	 << " ray: " << ray << " m_n: " << m_n
		  //	 << " q: " << q << " coord: " << coord 
		  //	 << " flb: " << errFun(lb) << " fub: " << errFun(ub)
		  //	 << endl;
		  //pthread_mutex_unlock(&printMutex);

		 // if (refraxInit != NULL) {
			//pthread_mutex_lock(&printMutex);
			//cerr << "[Ccamera::project] Init est error: "
		 // 		 << fabs(y - y0) << ", default est error: " 
		 // 		 << fabs(yObj / depth - y) << endl;
			//pthread_mutex_unlock(&printMutex);
		 // }

		  if (optRes != 1) {
			pthread_mutex_lock(&printMutex);
			cerr << "[Ccamera::project] Refraction optimization failed! Code = " 
				 << optRes 
				 << endl
				 << "lb: " << lb << " ub: " << ub << " y: " << y 
				 << " yObj: " << yObj << " y0: " << y0 
				 << " y0Obj: " << y0Obj << " depth: " << depth 
				 << " ray: " << ray << " m_n: " << m_n
				 << " q: " << q << " coord: " << coord 
				 << " flb: " << errFun(lb) << " fub: " << errFun(ub)
				 << endl;
			pthread_mutex_unlock(&printMutex);
		  }
		}
		else {
		  y = (lb + ub) / 2.0f;
		}
	  }

	  // Compute 3D coordinates of intersection with refraction interface
	  vtmp = m_t - m_n * m_d[0] + yDir * y * m_d[0];

	  q[0] = vtmp[0];
	  q[1] = vtmp[1];
	  q[2] = vtmp[2];
	} // end if (depth >= m_sumD)
	else if (depth < 0) {
	  // Ray is going away from refractive interface, no refraction.
	  q = coord;
	}
	else {
	  // Error, point is somewhere between refractive interfaces
	  q[0] = q[1] = q[2] = q[3] = 0.0;
	}
  }
#endif

  if (refrax != NULL)
	*refrax = q;

  // Perspective projection
  // Actually m_projection could represent other types of projections,
  // which would be bad because we assume perspective when computing
  // the refraction stuff above
  vtmp = standardProjection(q, level);

  return vtmp;
};

Vec4f Ccamera::unproject(const Vec3f& icoord, const int m_level) const {
  Mat3 A;
  Vec3 b(icoord[0], icoord[1], icoord[2]);
  for (int y = 0; y < 3; ++y) {
    for (int x = 0; x < 3; ++x)
      A[y][x] = m_projection[m_level][y][x];
    b[y] -= m_projection[m_level][y][3];    
  }
  Mat3 IA;
  invert(IA, A);
  Vec3 x = IA * b;
  return Vec4f(x, 1.0f);
}

void Ccamera::unprojectPath(const Vec3f& icoord, std::vector<Vec3f>& rayDir,
						    std::vector<Vec3f>& rayOrigin, const int level) const {
  if (m_txtType != 4) {
    cerr << "unprojectPath not supported for txtType: " << m_txtType << endl;
    exit (1);
  }

  double a, b, cosI, sinT2;

  Vec3 p = m_t;

  Vec3 v = m_iK[level] * icoord; // Unproject
  
  if (v[2] > 0) {
	v *= -1; // Make sure direction is away from camera
  }
  unitize(v);
  v = m_iR * v; // Rotate into world coordinate system

  // Origin and direction of first ray
  rayOrigin.push_back(p);
  rayDir.push_back(v);

  //pthread_mutex_lock(&printMutex);
  //cerr << "init v= " << rayDir[0] << ", p= " << rayOrigin[0] << endl;
  //pthread_mutex_unlock(&printMutex);

  for (int i = 0; i < m_numRefractions; i++) {
	// Refraction angle values
	// See Bram de Greve, "Reflections and Refractions in Ray Tracing"
	a = m_mu[i] / m_mu[i+1];
	cosI = -(v * m_n);
	sinT2 = a * a * (1.0 - cosI * cosI);

	if (sinT2 > 1.0) {
	  cerr << "[Ccamera::unprojectPath] Error: total internal reflection" << endl;
	}

	b = a * cosI - sqrt(1.0 - sinT2);

	// Intersection of ray with the next refractive interface
	p = p + (m_d[i] / cosI) * v;

	// Next ray direction
	v = a * v + b * m_n;

	rayOrigin.push_back(p);
	rayDir.push_back(v);

	//pthread_mutex_lock(&printMutex);
	//cerr << "iter " << i << ", v= " << rayDir[i+1] << ", p= " << rayOrigin[i+1] << endl;
	//pthread_mutex_unlock(&printMutex);
  }
}

double Ccamera::unproject2D(double depth, double y) const
{
	Vec2 v, p;
	double a, sinT2, norm;

	if (m_numRefractions <= 0) {
		// No refractions, directly back project
		return y * depth;
	}

	norm = sqrt(1.0 + y * y);
	v[0] = 1.0 / norm;
	v[1] = y / norm;

	p[0] = p[1] = 0.0;

	for (int i = 0; i < m_numRefractions; i++) {
		// Optimized 2D version of code in sfm_refractiveBackProject
		a = m_mu[i] / m_mu[i+1];
		sinT2 = a * a * (1.0 - v[0] * v[0]);

		if (sinT2 > 1.0) {
			printf("[sfm_backProjectObj2D] Error: total internal reflection\n");
		}

		// Intersection of ray with the next refractive interface
		p[0] += m_d[i];
		p[1] += (m_d[i] / v[0]) * v[1];

		v[0] = sqrt(1.0 - sinT2);
		v[1] *= a;
	}

	return p[1] + v[1] * (depth - p[0]) / v[0];
}