#ifndef IMAGE_CAMERA_H
#define IMAGE_CAMERA_H

#include <vector>
#include <string>
#include <climits>
#include "../numeric/vec4.h"
#include "../numeric/mat4.h"
#include "../numeric/mat3.h"

namespace Image {

class Ccamera {
 public:
  Ccamera(void);
  virtual ~Ccamera();
  
  // Update projection matrices from intrinsics and extrinsics
  void updateProjection(void);
  // Update all the camera related parameters
  void updateCamera(void);

  virtual void init(const std::string cname, const int maxLevel);
  void write(const std::string file);
  
  /*
   * Project 3D point coord to image coordinates.
   * Cheirality check: if point is behind camera, the returned vector
   * is (NaN, NaN, -1.0).
   *
   * Optional parameter refrax to save the projected point on the
   * refraction plane closest to the camera
   *
   * Optional refraxInit as starting guess
   *
   * refraxInit is read before refrax is assigned, so it is fine if 
   * they both point to the same variable.
   */
  Vec3f project(const Vec4f& coord, const int level, 
	            Vec4f *refrax = NULL, const Vec4f *refraxInit = NULL) const;

  inline Vec3f mult(const Vec4f& coord, const int level) const;
  
  static void setProjection(const std::vector<float>& intrinsics,
			    const std::vector<float>& extrinsics,
			    std::vector<Vec4f>& projection,
			    const int txtType);
  
  float getScale(const Vec4f& coord, const int level) const;
  void getPAxes(const Vec4f& coord, const Vec4f& normal,
		Vec4f& pxaxis, Vec4f& pyaxis, const int level = 0) const;

  void setAxesScale(const float axesScale);
  
  float getFocalX(const int level) const;
  float getFocalY(const int level) const;

  static void proj2q(Mat4& mat, double q[6]);
  static void q2proj(const double q[6], Mat4& mat);  
  static void setProjectionSub(double params[], std::vector<Vec4f>& projection,
			       const int level);

  float computeDistance(const Vec4f& point) const;
  float computeDepth(const Vec4f& point) const;  
  float computeDepthDif(const Vec4f& rhs, const Vec4f& lhs) const;

  //// Compute where the viewing ray passing through coord intersects
  //// with the plane abcd.
  //Vec4f intersect(const Vec4f& coord, const Vec4f& abcd) const;
  //void intersect(const Vec4f& coord, const Vec4f& abcd,
  //               Vec4f& cross, float& distance) const;

  // Computer a 3D coordinate that projects to a given image
  // coordinate. You can specify a different depth by the third
  // component of icoord.
  Vec4f unproject(const Vec3f& icoord, const int m_level) const;

  /**
	* Compute the light path corresponding to an image point
	* icoord. For a regular camera this is just a single ray
	* direction with origin corresponding to the camera center.
	* For a camera looking through flat refractive interfaces,
	* the direction for each refracted segment is given as well
	* as the intersection point on each interface. All
	* coordinates and directions are in the world coordinate 
	* system.
	*
	* Results are pushed onto the given vector containers, in
	* order from the camera center outwards. For efficiency,
	* ensure that the containers have sufficient capacity
	* (number of refractions plus 1).
	*/
  void unprojectPath(const Vec3f& icoord, std::vector<Vec3f>& rayDir,
					 std::vector<Vec3f>& rayOrigin, const int level) const;

  void setK(Mat3f& K) const;
  void setRT(Mat4f& RT) const;

  void getR(Mat3f& R) const;
  
  //----------------------------------------------------------------------
  // txt file name
  std::string m_cname;  
  // Optical center
  Vec4f m_center;
  // Optical axis
  Vec4f m_oaxis;
  
  float m_ipscale;
  // 3x4 projection matrix
  std::vector<std::vector<Vec4f> > m_projection;
  Vec3f m_xaxis;
  Vec3f m_yaxis;
  Vec3f m_zaxis;

  // intrinsic and extrinsic camera parameters. Compact form.
  std::vector<float> m_intrinsics;
  std::vector<float> m_extrinsics;
  // camera parameter type
  int m_txtType;

  // extrinsic parameters in full form
  // Transformation from world to camera coordinates is:
  //	x = m_R * (X - m_t)
  // i.e. m_t gives the camera center in world coordinates.
  // We should have that:
  //    P = m_K [m_R | -m_R*m_t]
  Mat3 m_R;
  Vec3 m_t;
  Mat3 m_iR; // inverse

  // intrinsic parameters in full form, one for each level
  std::vector<Mat3f> m_K;
  std::vector<Mat3f> m_iK; // inverse

  // flat refraction parameters.
  Vec3 m_n; // normal direction of refractive planes, in world coordinates
  int m_numRefractions;
  std::vector<float> m_d; // refractive layer thicknesses. Layer zero is the
						  // one containing the camera.
  float m_sumD;
  std::vector<float> m_mu; // refractive index of each layer.

 protected:
  int m_maxLevel;

  float m_axesScale;

  double m_refractEpsilon;

  Vec4f getOpticalCenter(void) const;

  struct RefProjectErrFun {
	const Ccamera *cam;
	double depth, yObj;

	RefProjectErrFun(const Ccamera *cam, double depth, double yObj) 
	  : cam(cam), depth(depth), yObj(yObj) { }

	double operator()(const double y) {
	  return yObj - cam->unproject2D(depth, y);
	  /*double err = yObj - cam->unproject2D(depth, y);
	  std::cerr << "Err: " << err << std::endl;
	  return err;*/
	}
  };

  /*
   * Back-project ray in 2D. This is an internal function used inside an
   * optimization procedure to compute refractive projection.
   *
   * Refractive interfaces are parallel to the Y axis, camera is looking
   * along the X axis.
   *
   *	depth object depth (must be positive, greater than sum(m_d))
   *	y y-coordinate of point to back project, at a distance 1 away from 
   *		the camera.
   */
  double unproject2D(double depth, double y) const;

  inline Vec3 standardProjection(const Vec4& q, const int level) const;
};

inline Vec3f Ccamera::mult(const Vec4f& coord,
			      const int level) const {
  Vec3f vtmp;    
  for (int i = 0; i < 3; ++i)
    vtmp[i] = m_projection[level][i] * coord;
  
  return vtmp;
};
 
inline Vec3 Ccamera::standardProjection(const Vec4& q, const int level) const {
  Vec3 vtmp;

  for (int i = 0; i < 3; ++i)
    vtmp[i] = (Vec4)m_projection[level][i] * q;

  if (vtmp[2] <= 0.0) {
    vtmp[0] = -0xffff;
    vtmp[1] = -0xffff;
    vtmp[2] = -1.0f;
    return vtmp;
  }
  else {
    vtmp /= vtmp[2];
  }
  
  vtmp[0] = std::max((double)(INT_MIN + 3.0),
				      std::min((double)(INT_MAX - 3.0), vtmp[0]));

  vtmp[1] = std::max((double)(INT_MIN + 3.0),
					  std::min((double)(INT_MAX - 3.0), vtmp[1]));
  
  return vtmp;
}

template<class T>
float computeEPD(const TMat3<T>& F, const TVec3<T>& p0, const TVec3<T>& p1) {
  TVec3<T> line = F * p1;
  const T ftmp = sqrt(line[0] * line[0] + line[1] * line[1]);
  if (ftmp == 0.0)
    return 0.0;

  line /= ftmp;
  return fabs(line * p0);
};

template<class T>
void setF(const Image::Ccamera& lhs, const Image::Ccamera& rhs,
	  TMat3<T>& F, const int level = 0) {
  const TVec4<T>& p00 = lhs.m_projection[level][0];
  const TVec4<T>& p01 = lhs.m_projection[level][1];
  const TVec4<T>& p02 = lhs.m_projection[level][2];

  const TVec4<T>& p10 = rhs.m_projection[level][0];
  const TVec4<T>& p11 = rhs.m_projection[level][1];
  const TVec4<T>& p12 = rhs.m_projection[level][2];

  F[0][0] = det(TMat4<T>(p01, p02, p11, p12));
  F[0][1] = det(TMat4<T>(p01, p02, p12, p10));
  F[0][2] = det(TMat4<T>(p01, p02, p10, p11));

  F[1][0] = det(TMat4<T>(p02, p00, p11, p12));
  F[1][1] = det(TMat4<T>(p02, p00, p12, p10));
  F[1][2] = det(TMat4<T>(p02, p00, p10, p11));

  F[2][0] = det(TMat4<T>(p00, p01, p11, p12));
  F[2][1] = det(TMat4<T>(p00, p01, p12, p10));
  F[2][2] = det(TMat4<T>(p00, p01, p10, p11));
 };
 
}; // namespace image

#endif // CAMERA_H
