#ifndef UTIL_H
#define UTIL_H

#include <cmath>
#include <algorithm>

namespace Image {
  /*
   * General-purpose root bracketing function
   * fun must define:
   *	float operator()(const float x)
   *
   * Return codes:
   *	1  Solution found
   *   -6  No sign change
   */
  template <class Functor>
  int fzero(Functor fun, double a, double b, double tolerance, double *ans)
  {
    int ret = 1;
	double c;

	// Optimization code based on Matlab's fzero function
	double fa, fb, fc; // function evaluated at a, b, c resp.
	double d; // update amount for b
	double e; // See below...
	double m; // half of the distance from b to c
	double s; // factor of change from previous step
	double p, q, r;
	double toler;

	fa = fun(a);
	fb = fun(b);

	if ((fa > 0 && fb > 0) || (fa < 0 && fb < 0)) {
	  ret = -6;
	  return ret;
	}

	fc = fb;

	int funcCount = 2;
	
	while (fb != 0.0 && a != b)
	{	
	  // b is the best result so far, a is the previous value of b, and
	  // c is on the opposite side of the zero from b
	  // c is always the current or a previous value of a
	  if ((fb > 0.0) == (fc > 0.0))
	  {
	    // b and c are on the same side of the zero, but previously they
		// were not. So a is on the other side of the current b, and set
		// c to the value of a
		c = a; fc = fa;
			
		d = b - a; e = d; // this is the step that b took in the last iteration
	  }
	  if (fabs(fc) < fabs(fb))
	  {
	    // latest b is worse than the result on the other side, so
		// swap b and c. 
		a = b;   b = c;   c = a;
		fa = fb; fb = fc; fc = fa;
	  }
		
	  m = 0.5 * (c - b);
	  toler = 2.0 * tolerance * max(fabs(b), 1.0);
		
	  if (fabs(m) <= toler || fb == 0.0) {
	    ret = 1;
		break;
	  }
			
	  if (fabs(e) < toler || fabs(fa) <= fabs(fb))
	  {
	    // e is either the step that b took in the last iteration, or,
		// if the last iteration was an interpolation, the step that b took
		// in the iteration before last. 
		// So that means we will only switch to bisection two steps after a 
		// small interpolation step, or if the solution got worse
			
		// Bisection
		d = m; e = m;
	  }
	  else
	  {
		s = fb / fa;
			
		if (a == c)
		{
		  // Linear interpolation
		  p = 2.0 * m * s;
		  q = 1.0 - s;
		}
		else 
		{
		  // Inverse quadratic interpolation
		  q = fa / fc;
		  r = fb / fc;
		  p = s * (2.0 * m * q * (q - r) - (b - a) * (r - 1.0));
		  q = (q - 1.0) * (r - 1.0) * (s - 1.0);
		}
		if (p > 0) q = -q; else p = -p;
			
		// Not sure what kind of test this is
		if ((2.0 * p < 3.0 * m * q - fabs(toler * q)) && (p < fabs(0.5 * e * q)))
		{
		  e = d; d = p/q;
		}
		else
		{
		  d = m; e = m;
		}
	  }
		
	  a = b;
	  fa = fb;
		
	  if (fabs(d) > toler)
	    b = b + d;
	  else if (b > c)
		b = b - toler;
	  else
		b = b + toler;
			
	  fb = fun(b);
		
	  funcCount++;
	}
	
	*ans = b;

	return ret;
  }

};

#endif