#include <time.h>
#include <numeric>
#include <algorithm>
#include <ctime>
#include <cmath>
#include <time.h>
#include <vector>
#include <cstring>
#include "seed.h"
#include "findMatch.h"
#include "lmmin.h"

using namespace Image;
using namespace PMVS3;
using namespace Patch;
using namespace std;

pthread_mutex_t cseedPrintMutex = PTHREAD_MUTEX_INITIALIZER;

Cseed::Cseed(CfindMatch& findMatch) : m_fm(findMatch) {
}

void Cseed::init(const std::vector<std::vector<Cpoint> >& points) {
  m_ppoints.clear();
  m_ppoints.resize(m_fm.m_num);

  for (int index = 0; index < m_fm.m_num; ++index) {
    const int gheight = m_fm.m_pos.m_gheights[index];
    const int gwidth = m_fm.m_pos.m_gwidths[index];
    m_ppoints[index].resize(gwidth * gheight);
  }

  readPoints(points);
}

void Cseed::readPoints(const std::vector<std::vector<Cpoint> >& points) {
  for (int index = 0; index < m_fm.m_num; ++index) {
    for (int i = 0; i < (int)points[index].size(); ++i) {
      Ppoint ppoint(new Cpoint(points[index][i]));
      ppoint->m_itmp = index;
      const int ix = ((int)floor(ppoint->m_icoord[0] + 0.5f)) / m_fm.m_csize;
      const int iy = ((int)floor(ppoint->m_icoord[1] + 0.5f)) / m_fm.m_csize;
      const int index2 = iy * m_fm.m_pos.m_gwidths[index] + ix;
      m_ppoints[index][index2].push_back(ppoint);
    }
  }
}

void Cseed::run(void) {
  m_fm.m_count = 0;
  m_fm.m_jobs.clear();
  m_scounts.resize(m_fm.m_CPU);
  m_fcounts0.resize(m_fm.m_CPU);
  m_fcounts1.resize(m_fm.m_CPU);
  m_pcounts.resize(m_fm.m_CPU);
  fill(m_scounts.begin(), m_scounts.end(), 0);
  fill(m_fcounts0.begin(), m_fcounts0.end(), 0);
  fill(m_fcounts1.begin(), m_fcounts1.end(), 0);
  fill(m_pcounts.begin(), m_pcounts.end(), 0);
  
  vector<int> vitmp;
  for (int i = 0; i < m_fm.m_tnum; ++i)
    vitmp.push_back(i);

  random_shuffle(vitmp.begin(), vitmp.end());
  m_fm.m_jobs.insert(m_fm.m_jobs.end(), vitmp.begin(), vitmp.end());

  cerr << "[Cseed::run] adding seeds " << endl;
  
  m_fm.m_pos.clearCounts();

  // If there already exists a patch, don't use
  for (int index = 0; index < (int)m_fm.m_tnum; ++index) {
    for (int j = 0; j < (int)m_fm.m_pos.m_pgrids[index].size(); ++j) {
      if (!m_fm.m_pos.m_pgrids[index][j].empty())
        m_fm.m_pos.m_counts[index][j] = m_fm.m_countThreshold2;
    }
  }

  time_t curtime = time(NULL);
  vector<pthread_t> threads(m_fm.m_CPU);
  for (int i = 0; i < m_fm.m_CPU; ++i)
    pthread_create(&threads[i], NULL, initialMatchThreadTmp, (void*)this);
  for (int i = 0; i < m_fm.m_CPU; ++i)
    pthread_join(threads[i], NULL);
  //----------------------------------------------------------------------
  cerr << "done" << endl;
  cerr << "---- Initial: " << (time(NULL) - curtime)/CLOCKS_PER_SEC << " secs ----" << endl;

  const int trial = accumulate(m_scounts.begin(), m_scounts.end(), 0);
  const int fail0 = accumulate(m_fcounts0.begin(), m_fcounts0.end(), 0);
  const int fail1 = accumulate(m_fcounts1.begin(), m_fcounts1.end(), 0);
  const int pass = accumulate(m_pcounts.begin(), m_pcounts.end(), 0);
  cerr << "Total pass fail0 fail1 refinepatch: "
       << trial << ' ' << pass << ' '
       << fail0 << ' ' << fail1 << ' ' << pass + fail1 << endl;
  cerr << "Total pass fail0 fail1 refinepatch: "
       << 100 * trial / (float)trial << ' '
       << 100 * pass / (float)trial << ' '
       << 100 * fail0 / (float)trial << ' '
       << 100 * fail1 / (float)trial << ' '
       << 100 * (pass + fail1) / (float)trial << endl;
}

void Cseed::initialMatchThread(void) {
  pthread_rwlock_wrlock(&m_fm.m_lock);
  const int id = m_fm.m_count++;
  pthread_rwlock_unlock(&m_fm.m_lock);

  while (1) {
    int index = -1;
    pthread_rwlock_wrlock(&m_fm.m_lock);
    if (!m_fm.m_jobs.empty()) {
      index = m_fm.m_jobs.front();
      m_fm.m_jobs.pop_front();
    }
    pthread_rwlock_unlock(&m_fm.m_lock);
    if (index == -1)
      break;

    initialMatch(index, id);
  }
}

void* Cseed::initialMatchThreadTmp(void* arg) {
  ((Cseed*)arg)->initialMatchThread();
  return NULL;
}

void Cseed::clear(void) {
  vector<vector<vector<Ppoint> > >().swap(m_ppoints);
}

void Cseed::initialMatch(const int index, const int id) {
  vector<int> indexes;
  m_fm.m_optim.collectImages(index, indexes);

  if (m_fm.m_tau < (int)indexes.size())
    indexes.resize(m_fm.m_tau);
  
  if (indexes.empty())
    return;  

  int totalcount = 0;
  //======================================================================
  // for each feature point, starting from the optical center, keep on
  // matching until we find candidateThreshold patches
  const int gheight = m_fm.m_pos.m_gheights[index];
  const int gwidth = m_fm.m_pos.m_gwidths[index];

  int index2 = -1;
  for (int y = 0; y < gheight; ++y) {
    for (int x = 0; x < gwidth; ++x) {
      ++index2;
      if (!canAdd(index, x, y))
		continue;

      for (int p = 0; p < (int)m_ppoints[index][index2].size(); ++p) {
		// collect features that satisfies epipolar geometry
		// constraints and sort them according to the differences of
		// distances between two cameras.
		vector<Ppoint> vcp;
		collectCandidates(index, indexes,
							  *m_ppoints[index][index2][p], vcp);
        
		vector<Vec3f> rayDirs, rayOrigins;
		if (m_fm.m_pss.m_photos[index].m_numRefractions > 0) {
		  m_fm.m_pss.m_photos[index].unprojectPath(m_ppoints[index][index2][p]->m_icoord, 
			rayDirs, rayOrigins, m_fm.m_level);
		  //cerr << "New normal method!" << endl;
		}

		int count = 0;
		Cpatch bestpatch;
		//======================================================================
		for (int i = 0; i < (int)vcp.size(); ++i) {
		  Cpatch patch;
		  patch.m_coord = vcp[i]->m_coord;

#if 1
		  if (m_fm.m_pss.m_photos[index].m_numRefractions > 0) {
			patch.m_normal = Vec4f(-1 * rayDirs.back(), 0.0f);
		  }
		  else 
#endif
		  {
			patch.m_normal = m_fm.m_pss.m_photos[index].m_center - patch.m_coord;
		  }
		  unitize(patch.m_normal);
		  // HOW IS IT CORRECT TO DO THIS AFTER unitize?? Unless 4th entry already
		  // zero from the subtraction?
		  patch.m_normal[3] = 0.0; 
		  patch.m_flag = 0;
		  patch.m_generation = m_fm.getIterNum();

		  ++m_fm.m_pos.m_counts[index][index2];
		  const int ix = ((int)floor(vcp[i]->m_icoord[0] + 0.5f)) / m_fm.m_csize;
		  const int iy = ((int)floor(vcp[i]->m_icoord[1] + 0.5f)) / m_fm.m_csize;
		  const int index3 = iy * m_fm.m_pos.m_gwidths[vcp[i]->m_itmp] + ix;
		  if (vcp[i]->m_itmp < m_fm.m_tnum)
			++m_fm.m_pos.m_counts[vcp[i]->m_itmp][index3];
          
		  const int flag = initialMatchSub(index, vcp[i]->m_itmp, id, patch);
		  if (flag == 0) {
			++count;
			if (bestpatch.score(m_fm.m_nccThreshold) <
					patch.score(m_fm.m_nccThreshold))
			  bestpatch = patch;
			if (m_fm.m_countThreshold0 <= count)
			  break;
		  }
		}

		if (count != 0) {
		  Ppatch ppatch(new Cpatch(bestpatch));
		  m_fm.m_pos.addPatch(ppatch);
		  ++totalcount;
			  break;
		}
      }
    }
  }
}

//#define PRINT_CELLS_COUT

void Cseed::collectCells(const int index0, const int index1,
                         const Cpoint& p0, vector<Vec2i>& cells,
						 vector<Vec3f>& curvePts, 
						 vector<int>& cellCurveIdx) {
#ifdef PRINT_CELLS_COUT
  // We're gonna print the epipolar lines/curves for debugging.
  pthread_mutex_lock(&cseedPrintMutex);

  cout << index0 << " " << index1 << endl;
  cout << p0.m_icoord[0] << " " << p0.m_icoord[1] << endl;
#endif

  Vec3 point(p0.m_icoord[0], p0.m_icoord[1], p0.m_icoord[2]);
#ifdef DEBUG
  if (p0.m_icoord[2] != 1.0f) {
    cerr << "Impossible in collectCells" << endl;    exit (1);
  }
#endif
  
#if 1
  if (m_fm.m_pss.m_photos[index1].m_numRefractions > 0) {
	// Refraction, sample along epipolar curve

	float camDist = norm(m_fm.m_pss.m_photos[index1].m_t - m_fm.m_pss.m_photos[index0].m_t);
	float maxDepth = (camDist/2)/tan(m_fm.m_maxAngleThreshold/2);

	int nRef0 = m_fm.m_pss.m_photos[index0].m_numRefractions;
	int nRef1 = m_fm.m_pss.m_photos[index1].m_numRefractions;

	vector<Vec3f> rayDirs, rayOrigins;
	rayDirs.reserve(nRef0+1); rayOrigins.reserve(nRef0+1);
	Vec3f rayDir, rayOrigin;

	// Compute final refracted ray from first camera. Let's assume that the
	// object is positioned after all the refractions.
	m_fm.m_pss.m_photos[index0].unprojectPath(p0.m_icoord, rayDirs, 
	  rayOrigins, m_fm.m_level);

	rayDir = rayDirs.back();
	rayOrigin = rayOrigins.back();

	// We want to project points along the ray to the second camera, sampling
	// the curve as a piecewise linear approximation.
	// Sampling step size will be adjusted so that the angular change in the
	// second image is (very) rougly at a desired amount
	float stepAngle = 
	  atan(4 * m_fm.m_csize / m_fm.m_pss.m_photos[index1].getFocalX(m_fm.m_level));

	float alpha = 0;
	while (alpha < maxDepth) {
	  Vec3f pt = rayOrigin + alpha * rayDir;
	  Vec4f coord(pt[0], pt[1], pt[2], 1.0f);
	  Vec3f samplePt = m_fm.m_pss.m_photos[index1].project(coord, m_fm.m_level);

	  if (samplePt[2] == 1.0f) {
		curvePts.push_back(samplePt);
	  }

	  Vec3f camToPt = (Vec3)pt - m_fm.m_pss.m_photos[index1].m_t;
	  float ptDist = norm(camToPt);
	  float dot = camToPt/ptDist * rayDir;
	  float dirAngle = acos(dot);
	  if (dot > 0)
		dirAngle = M_PI - dirAngle;

	  alpha += ptDist * sin(stepAngle) / sin(dirAngle + stepAngle);
	}
  }
  else 
#endif
  {
	// No refraction, standard epipolar geometry

	// fundamental matrix...
	Mat3 F;
	Image::setF(m_fm.m_pss.m_photos[index0], m_fm.m_pss.m_photos[index1],
				F, m_fm.m_level);

	// epipolar line
	Vec3 line = transpose(F) * point;

	if (line[0] == 0.0 && line[1] == 0.0) {
	  cerr << "Point right on top of the epipole?"
		   << index0 << ' ' << index1 << endl;
	  return;
	}


	Vec3f lpa, lpb;

	if (fabs(line[0]) > fabs(line[1])) {
	  // epipolar line is more vertical than horizontal
	  lpa[1] = 0.0f;
	  lpb[1] = m_fm.m_pss.m_photos[index1].getHeight(m_fm.m_level) - 1;

	  lpa[0] = -line[2] / line[0];
	  lpb[0] = (-line[1] * lpb[1] - line[2]) / line[0];
	}
	else {
	  lpa[0] = 0.0f;
	  lpb[0] = m_fm.m_pss.m_photos[index1].getWidth(m_fm.m_level) - 1;

	  lpa[1] = -line[2] / line[1];
	  lpb[1] = (-line[0] * lpb[0] - line[2]) / line[1];
	}
	lpa[2] = lpb[2] = 1.0f;

	curvePts.push_back(lpa); curvePts.push_back(lpb);
  }

  collectCellsAlongLine(index1, curvePts, cells, cellCurveIdx);

#ifdef PRINT_CELLS_COUT
  cout << curvePts.size() << endl;

  for (unsigned int i = 0; i < curvePts.size(); i++) {
	cout << curvePts[i][0] * (0x1 << m_fm.m_level) << " " 
	     << curvePts[i][1] * (0x1 << m_fm.m_level) << endl;
  }

  cout << cells.size() << " " << m_fm.m_csize * (0x1 << m_fm.m_level) << endl;

  for (unsigned int i = 0; i < cells.size(); i++) {
	cout << cells[i][0] << " " << cells[i][1] << endl;
  }

  pthread_mutex_unlock(&cseedPrintMutex);
#endif
}

void Cseed::collectCellsAlongLine(int index, const std::vector<Vec3f>& pts, 
								  std::vector<Vec2i>& cells,
								  std::vector<int>& ptIdx) {
  const int gWidth = m_fm.m_pos.m_gwidths[index];
  const int gHeight = m_fm.m_pos.m_gheights[index];
  const int iWidth = m_fm.m_pss.m_photos[index].getWidth(m_fm.m_level);
  const int iHeight = m_fm.m_pss.m_photos[index].getHeight(m_fm.m_level);
  const int numPts = pts.size();

  // To avoid adding duplicates.
  bool *cellAdded = new bool[gWidth * gHeight];
  memset(cellAdded, 0, gWidth * gHeight * sizeof(bool));

  bool lastRound = false;

  // Debugging
  int numCellsTouched = 0;

  for (int i = 0; i < numPts-1; i++) {
	Vec3f pt1 = pts[i], pt2 = pts[i+1];
	Vec3f line = cross(pt1, pt2);

	// Optimization: both points are outside image and the line segment never crosses into the image, skip it entirely
	if (((pt1[0] < 0) || (pt1[0] >= iWidth) || (pt1[1] < 0) || (pt1[1] >= iHeight)) &&
	    ((pt2[0] < 0) || (pt2[0] >= iWidth) || (pt2[1] < 0) || (pt2[1] >= iHeight))) {
	  // Check if it intersects any of the lines of the bounding rectangle
	  bool intersects = false;
	  float xa, xb, ya, yb;

	  if (pt1[0] < pt2[0]) { xa = pt1[0]; xb = pt2[0]; } else { xa = pt2[0]; xb = pt1[0]; }
	  if (pt1[1] < pt2[1]) { ya = pt1[1]; yb = pt2[1]; } else { ya = pt2[1]; yb = pt1[1]; }

	  if (line[1] != 0) {
		float interX0 = -line[2]/line[1], interXW = -(line[2] + line[0]*(iWidth-1))/line[1];
		intersects = ((ya <= interX0) && (interX0 <= yb)) || ((ya <= interXW) && (interXW <= yb));
	  }

	  if (!intersects && (line[0] != 0)) {
	    float interY0 = -line[2]/line[0], interYH = -(line[1]*(iHeight-1) + line[2])/line[0];
		intersects = ((xa <= interY0) && (interY0 <= xb)) || ((xa <= interYH) && (interYH <= xb));
	  }

	  if (!intersects)
		continue;
	}


	int gxStart = ((int)floor(pt1[0] + 0.5f)) / m_fm.m_csize;
	int gyStart = ((int)floor(pt1[1] + 0.5f)) / m_fm.m_csize;
	int gxEnd = ((int)floor(pt2[0] + 0.5f)) / m_fm.m_csize;
	int gyEnd = ((int)floor(pt2[1] + 0.5f)) / m_fm.m_csize;

	int gx = gxStart, gy = gyStart;

	while (true) {
	  numCellsTouched++;

	  if (gx >= 0 && gx < gWidth && gy >= 0 && gy < gHeight) {
		int cellIdx = gy * gWidth + gx;

		// Add neighbouring cells in the four cardinal directions
		if (!cellAdded[cellIdx]) { 
		  cells.push_back(TVec2<int>(gx, gy)); ptIdx.push_back(i); cellAdded[cellIdx] = true; 
		}

		if ((gx + 1) < gWidth && !cellAdded[cellIdx + 1]) {
		  cells.push_back(TVec2<int>(gx + 1, gy)); ptIdx.push_back(i); cellAdded[cellIdx + 1] = true; 
		}

		if ((gx - 1) >= 0 && !cellAdded[cellIdx - 1]) {
		  cells.push_back(TVec2<int>(gx - 1, gy)); ptIdx.push_back(i); cellAdded[cellIdx - 1] = true; 
		}

		if ((gy + 1) < gHeight && !cellAdded[cellIdx + gWidth]) {
		  cells.push_back(TVec2<int>(gx, gy + 1)); ptIdx.push_back(i); cellAdded[cellIdx + gWidth] = true; 
		}

		if ((gy - 1) >= 0 && !cellAdded[cellIdx - gWidth]) {
		  cells.push_back(TVec2<int>(gx, gy - 1)); ptIdx.push_back(i); cellAdded[cellIdx - gWidth] = true; 
		}
	  }

	  if (lastRound)
		break;

	  // Advance to next cell
	  float fy, fx;
	  if (abs(gyEnd - gyStart) > abs(gxEnd - gxStart)) {
		// line is more vertical than horizontal
		gy += (gyEnd > gyStart)? 1 : -1;

		if (gy == gyEnd) {
		  if (i < numPts - 2)
			break;
		  else
			lastRound = true; // One more iteration to cap off end of line
		}

		fy = ((float)gy + 0.5f) * m_fm.m_csize - 0.5f;
		fx = (- line[1] * fy - line[2]) / line[0];
		gx = ((int)floor(fx + 0.5f)) / m_fm.m_csize;
	  }
	  else if (gxEnd != gxStart) {
		// line is more horizontal than vertical
		gx += (gxEnd > gxStart)? 1 : -1;

		if (gx == gxEnd) {
		  if (i < numPts - 2)
			break;
		  else
			lastRound = true;
		}

		fx = ((float)gx + 0.5f) * m_fm.m_csize - 0.5f;
		fy = (- line[0] * fx - line[2]) / line[1];
		gy = ((int)floor(fy + 0.5f)) / m_fm.m_csize;
	  }
	  else {
		// gxEnd == gxStart and gyEnd == gyStart
		// We have already added all relevant, surrounding cells
		break;
	  }
	}
  }

  delete[] cellAdded;

  //pthread_mutex_lock(&cseedPrintMutex);
  //cerr << "Cells touched " << numCellsTouched << "/" << cells.size() << " added. " << numPts-1 << " line segments." << endl;
  //pthread_mutex_unlock(&cseedPrintMutex);
}

float Cseed::localCurveDist(Vec3f coord, const std::vector<Vec3f>& curvePts, int ptIdx)
{
  // Compute minimum of distances from three adjacent curve segments near ptIdx
  float minDist = -1.0f;

  int i = max(0, ptIdx - 1);
  int last = min((int)curvePts.size() - 2, ptIdx + 1);

  for (; i <= last; i++) {
	Vec3f line = cross(curvePts[i], curvePts[i+1]);
	line /= sqrt(line[0]*line[0] + line[1]*line[1]);

	// Orthogonal projection of the point might not lie between the segment endpoints
	// but let's not worry about that for now. If the curve is sufficiently straight
	// this should be a good approximation
	float dist = fabs(line * coord);

	if ((dist < minDist) || (minDist < 0))
	  minDist = dist;
  }

  return minDist;
}

// make sorted array of feature points in images, that satisfy the
// epipolar geometry coming from point in image
void Cseed::collectCandidates(const int index, const std::vector<int>& indexes,
                              const Cpoint& point, std::vector<Ppoint>& vcp) 
{
  // Point in first image
  const Vec3f p0(point.m_icoord[0], point.m_icoord[1], 1.0f);

  for (int i = 0; i < (int)indexes.size(); ++i) {        
    const int indexid = indexes[i];
    
    vector<TVec2<int> > cells;
	vector<Vec3f> curvePts;
	vector<int> cellCurveIdx;
    collectCells(index, indexid, point, cells, curvePts, cellCurveIdx);

	// Compute fundamental matrix for use in computeEPD below
    Mat3f F;
    Image::setF(m_fm.m_pss.m_photos[index], m_fm.m_pss.m_photos[indexid],
                F, m_fm.m_level);
    
    for (int i = 0; i < (int)cells.size(); ++i) {
      const int x = cells[i][0];      const int y = cells[i][1];

      if (!canAdd(indexid, x, y))
		continue;

      const int index2 = y * m_fm.m_pos.m_gwidths[indexid] + x;

      vector<Ppoint>::iterator begin = m_ppoints[indexid][index2].begin();
      vector<Ppoint>::iterator end = m_ppoints[indexid][index2].end();

      while (begin != end) {
        Cpoint& rhs = **begin;
        // ? use type to reject candidates?
        if (point.m_type != rhs.m_type) {
          ++begin; // skip point
          continue;
        }
          
		// Point in second image
        const Vec3f p1(rhs.m_icoord[0], rhs.m_icoord[1], 1.0f);

		// Check epipolar distance
//        if (m_fm.m_epThreshold <= Image::computeEPD(F, p0, p1)) {
		float curveDist = localCurveDist(p1, curvePts, cellCurveIdx[i]);
		//pthread_mutex_lock(&cseedPrintMutex);
		//cerr << "Point " << p1 << " curve dist " << curveDist
		//	 << " EPD " << Image::computeEPD(F, p0, p1) << endl;
		//pthread_mutex_unlock(&cseedPrintMutex);
		if (curveDist < 0 || curveDist >= m_fm.m_epThreshold) {
          ++begin; // skip point         
          continue;
        }
        vcp.push_back(*begin);
        ++begin;
      }
    }
  }
  
  // set distances to m_response
  vector<Ppoint> vcptmp;
  for (int i = 0; i < (int)vcp.size(); ++i) {
	// Triangulate 3D point
    unproject(index, vcp[i]->m_itmp, point, *vcp[i], vcp[i]->m_coord);
    
	Vec3f reproj = m_fm.m_pss.m_photos[index].project(vcp[i]->m_coord, 
													  m_fm.m_level);

	// Cheirality check with first camera
    /*if (m_fm.m_pss.m_photos[index].m_projection[m_fm.m_level][2] *
        vcp[i]->m_coord <= 0.0)
      continue;*/
	if (reproj[2] <= 0.0f)
	  continue;

	// Check that reprojected point is not masked out and is within
	// image bounds
	// TODO duplicate project() calls in here...
    if (m_fm.m_pss.getMask(vcp[i]->m_coord, m_fm.m_level) == 0 ||
        m_fm.insideBimages(vcp[i]->m_coord) == 0)
      continue;

    // ??? difference in distance from the two cameras ???
    vcp[i]->m_response =
      fabs(norm(vcp[i]->m_coord - m_fm.m_pss.m_photos[index].m_center) -
           norm(vcp[i]->m_coord - m_fm.m_pss.m_photos[vcp[i]->m_itmp].m_center));
    
    vcptmp.push_back(vcp[i]);
  }
  vcptmp.swap(vcp);
  sort(vcp.begin(), vcp.end()); // sort by m_response
}

int Cseed::canAdd(const int index, const int x, const int y) {
  if (!m_fm.m_pss.getMask(index, m_fm.m_csize * x, m_fm.m_csize * y, m_fm.m_level))
    return 0;

  const int index2 = y * m_fm.m_pos.m_gwidths[index] + x;

  if (m_fm.m_tnum <= index)
    return 1;
  
  // Check if m_pgrids already contains something
  if (!m_fm.m_pos.m_pgrids[index][index2].empty())
    return 0;

  //??? critical
  if (m_fm.m_countThreshold2 <= m_fm.m_pos.m_counts[index][index2])
    return 0;
  
  return 1;
}

void Cseed::unproject(const int index0, const int index1,
                      const Cpoint& p0, const Cpoint& p1,
                      Vec4f& coord) const {
						
  if (m_fm.m_pss.m_photos[index0].m_numRefractions > 0 &&
	  m_fm.m_pss.m_photos[index1].m_numRefractions > 0) {

	vector<Vec3f> dirs0, pos0, dirs1, pos1;

	dirs0.reserve(m_fm.m_pss.m_photos[index0].m_numRefractions + 1);
	pos0.reserve(m_fm.m_pss.m_photos[index0].m_numRefractions + 1);
	dirs1.reserve(m_fm.m_pss.m_photos[index1].m_numRefractions + 1);
	pos1.reserve(m_fm.m_pss.m_photos[index1].m_numRefractions + 1);

	m_fm.m_pss.m_photos[index0].unprojectPath(p0.m_icoord, dirs0, pos0, m_fm.m_level);
	m_fm.m_pss.m_photos[index1].unprojectPath(p1.m_icoord, dirs1, pos1, m_fm.m_level);

	midpoint_triangulate(pos0.back(), pos1.back(), dirs0.back(), dirs1.back(), coord);

	Vec4f initRefraxL(pos0[1], 1.0f), initRefraxR(pos1[1], 1.0f);
	polishTriangulation(coord, index0, index1, p0.m_icoord, p1.m_icoord, 
	  &initRefraxL, &initRefraxR);
  }
  else if ((m_fm.m_pss.m_photos[index0].m_numRefractions > 0) !=
		   (m_fm.m_pss.m_photos[index1].m_numRefractions > 0)) {
    cerr << "[Cseed::unproject] Mixed refractive and non-refractive cameras not supported." << endl;
    exit (1);
  }
  else {
	Mat4 A;
	A[0][0] =
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][0][0] -
	  p0.m_icoord[0] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][0];
	A[0][1] =
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][0][1] -
	  p0.m_icoord[0] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][1];
	A[0][2] =
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][0][2] -
	  p0.m_icoord[0] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][2];
	A[1][0] =
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][1][0] -
	  p0.m_icoord[1] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][0];
	A[1][1] =
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][1][1] -
	  p0.m_icoord[1] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][1];
	A[1][2] =
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][1][2] -
	  p0.m_icoord[1] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][2];
	A[2][0] =
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][0][0] -
	  p1.m_icoord[0] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][0];
	A[2][1] =
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][0][1] -
	  p1.m_icoord[0] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][1];
	A[2][2] =
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][0][2] -
	  p1.m_icoord[0] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][2];
	A[3][0] =
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][1][0] -
	  p1.m_icoord[1] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][0];
	A[3][1] =
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][1][1] -
	  p1.m_icoord[1] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][1];
	A[3][2] =
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][1][2] -
	  p1.m_icoord[1] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][2];

	Vec4 b;
	b[0] =
	  p0.m_icoord[0] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][3] -
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][0][3];
	b[1] =
	  p0.m_icoord[1] * m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][2][3] -
	  m_fm.m_pss.m_photos[index0].m_projection[m_fm.m_level][1][3];
	b[2] =
	  p1.m_icoord[0] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][3] -
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][0][3];
	b[3] =
	  p1.m_icoord[1] * m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][2][3] -
	  m_fm.m_pss.m_photos[index1].m_projection[m_fm.m_level][1][3];

	Mat4 AT = transpose(A);
	Mat4 ATA = AT * A;
	Vec4 ATb = AT * b;

	Mat3 ATA3;
	for (int y = 0; y < 3; ++y)
	  for (int x = 0; x < 3; ++x)
		ATA3[y][x] = ATA[y][x];
	Vec3 ATb3;
	for (int y = 0; y < 3; ++y)
	  ATb3[y] = ATb[y];
  
	Mat3 iATA3;
	invert(iATA3, ATA3);
	Vec3 ans = iATA3 * ATb3;
	for (int y = 0; y < 3; ++y)
	  coord[y] = ans[y];
	coord[3] = 1.0f;
  }
}		       

void Cseed::midpoint_triangulate(const Vec3f& orgl, const Vec3f& orgr,
								 const Vec3f & dirl, const Vec3f & dirr, 
								 Vec4f& coord) const {
  Mat3f A,iA;
  Vec3f b;
  Vec3f n;
  Vec3f stw;
  Vec3f Xl,Xr,X;

  n=dirl^dirr;
  b=orgr-orgl;
  A[0]=dirl;
  A[1]=-dirr;
  A[2]=n;
  A=transpose(A);
  invert(iA,A);
  stw=iA*b;

  Xl=orgl+stw[0]*dirl;
  Xr=orgr+stw[1]*dirr;
  X=(Xl+Xr)/2.0f;

  coord[0]=X[0];
  coord[1]=X[1];
  coord[2]=X[2];
  coord[3]=1;
}

void Cseed::polishTriangulation(Vec4f &coord, const int indexL, const int indexR,
							    const Vec3f &icoordL, const Vec3f &icoordR,
    	     				    const Vec4f *refraxInitL, const Vec4f *refraxInitR) const {

  PolishTriangulationLmData lmData(m_fm.m_pss.m_photos[indexL], m_fm.m_pss.m_photos[indexR]);

  lmData.initRefraxL = refraxInitL;
  lmData.initRefraxR = refraxInitR;
  lmData.icoordL = icoordL;
  lmData.icoordR = icoordR;
  lmData.level = m_fm.m_level;
  
  double x[3] = {coord[0], coord[1], coord[2]};
  
  lm_control_struct control = lm_control_float;
  control.gtol = 0;
  //control.epsilon = 1e-6;
  //control.printflags = 0;

  lm_status_struct status;

  lmmin(3, x, 4, (void *)&lmData, polishTriangulationErrFun, &control, &status, lm_printout_std);

  // status.info 0 to 3 are "good", the rest are bad
  if (status.info >= 0 && status.info <= 3) {
	coord[0] = x[0];
	coord[1] = x[1];
	coord[2] = x[2];
  }
  else {
	//cerr << "[Cseed::polishTriangulation] Optimization failed (reason " << status.info << "). leaving coord unchanged." << endl;
  }
}

void Cseed::polishTriangulationErrFun(const double *par, int m_dat, const void *data,
									  double *fvec, int *info) {
  const PolishTriangulationLmData *lmData = reinterpret_cast<const PolishTriangulationLmData *>(data);
  Vec4f coord(par[0], par[1], par[2], 1.0f);

  Vec3f projL = lmData->photoL.project(coord, lmData->level, NULL, lmData->initRefraxL);
  Vec3f projR = lmData->photoR.project(coord, lmData->level, NULL, lmData->initRefraxR);

  fvec[0] = projL[0] - lmData->icoordL[0];
  fvec[1] = projL[1] - lmData->icoordL[1];
  fvec[2] = projR[0] - lmData->icoordR[0];
  fvec[3] = projR[1] - lmData->icoordR[1];
}


// starting with (index, indexs), set visible images by looking at correlation.
int Cseed::initialMatchSub(const int index0, const int index1,
                           const int id, Cpatch& patch) {
  //----------------------------------------------------------------------
  patch.m_images.clear();
  patch.m_images.push_back(index0);
  patch.m_images.push_back(index1);

  ++m_scounts[id];

  //----------------------------------------------------------------------
  // We know that patch.m_coord is inside bimages and inside mask
  if (m_fm.m_optim.preProcess(patch, id, 1)) {
    ++m_fcounts0[id];
    return 1;
  }
  
  //----------------------------------------------------------------------  
  m_fm.m_optim.refinePatch(patch, id, 100);

  //----------------------------------------------------------------------
  if (m_fm.m_optim.postProcess(patch, id, 1)) {
    ++m_fcounts1[id];
    return 1;
  }
  
  ++m_pcounts[id];
  //----------------------------------------------------------------------
  return 0;
}
