#ifndef PMVS3_SEED_H
#define PMVS3_SEED_H

#include <boost/shared_ptr.hpp>
#include <vector>
#include "patch.h"
#include "point.h"
#include "../image/photo.h"

#include <pthread.h>

namespace PMVS3 {
class CfindMatch;
typedef boost::shared_ptr<Cpoint> Ppoint;
  
class Cseed {
 public:
  Cseed(CfindMatch& findMatch);
  virtual ~Cseed() {};

  void init(const std::vector<std::vector<Cpoint> >& points);
  void run(void);
  void clear(void);

 protected:
  void readPoints(const std::vector<std::vector<Cpoint> >& points);
  int canAdd(const int index, const int x, const int y);  

  void initialMatch(const int index, const int id);

  /*
   * If refractive projection is used in camera 1 (observing the back-
   * projected ray), then curvePts will be populated with a list of 
   * image coordinates where the curved image of the ray is sampled.
   * cellCurveIdx will be populated with one entry for each cell giving
   * the index of a curve point close to the center of that cell.
   *
   * If there is no refraction, then curvePts will just contain the 
   * endpoints of the epipolar line.
   */
  void collectCells(const int index0, const int index1,
                    const Cpoint& p0, std::vector<Vec2i>& cells,
					std::vector<Vec3f>& curvePts, 
					std::vector<int>& cellCurveIdx);
  
  /*
   * Collect all cells close to the set of line segments defined by
   * the list of points pts. No duplicates will be returned.
   * Each cell has a corresponding entry in ptIdx giving the index of 
   * the point that caused it to be added.
   */
  void collectCellsAlongLine(int index, const std::vector<Vec3f>& pts, 
							 std::vector<Vec2i>& cells,
							 std::vector<int>& ptIdx);

  float localCurveDist(Vec3f coord, const std::vector<Vec3f>& curvePts, 
					   int ptIdx);

  void collectCandidates(const int index, const std::vector<int>& indexes,
                         const Cpoint& point, std::vector<Ppoint>& vcp);

  int initialMatchSub(const int index0, const int index1,
                      const int id, Patch::Cpatch& patch);
  
  void unproject(const int index0, const int index1,
                 const Cpoint& p0, const Cpoint& p1,
                 Vec4f& coord) const;

  /*
   * orgl and orgr are the origin points of the left and right rays
   * dirl and dirr are the ray directions
   * Everything should be given in a single coordinate system
   */
  void midpoint_triangulate(const Vec3f& orgl, const Vec3f& orgr,
							const Vec3f & dirl, const Vec3f & dirr, 
							Vec4f& coord) const;

  /*
   * refraxInitL and refraxInitR can be null if there is no refraction.
   */
  void Cseed::polishTriangulation(Vec4f &coord, const int indexL, const int indexR,
	  						      const Vec3f &icoordL, const Vec3f &icoordR,
    	       				      const Vec4f *refraxInitL = NULL, const Vec4f *refraxInitR = NULL) const;

  struct PolishTriangulationLmData {
	const Image::Cphoto &photoL, &photoR;
    const Vec4f *initRefraxL, *initRefraxR;
	Vec3f icoordL, icoordR;
	int level;

	PolishTriangulationLmData(const Image::Cphoto &photoL, const Image::Cphoto &photoR) : photoL(photoL), photoR(photoR) {}
  };

  static void polishTriangulationErrFun(const double *par, int m_dat, const void *data,
									    double *fvec, int *info);

  //----------------------------------------------------------------------
  CfindMatch& m_fm;
  // points in a grid. For each index, grid
  std::vector<std::vector<std::vector<Ppoint> > > m_ppoints;

  //----------------------------------------------------------------------
  // thread related
  //----------------------------------------------------------------------
  void initialMatchThread(void);
  static void* initialMatchThreadTmp(void* arg);

  // Number of trials
  std::vector<int> m_scounts;
  // Number of failures in the prep
  std::vector<int> m_fcounts0;
  // Number of failures in the post processing
  std::vector<int> m_fcounts1;
  // Number passes
  std::vector<int> m_pcounts;
};
};

#endif // PMVS3_SEED_H
