/* 
 *  Copyright (c) 2008  Noah Snavely (snavely (at) cs.washington.edu)
 *    and the University of Washington
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 */

/* sfm-driver.h */
/* Driver for sfm routines */

#ifndef __sfm_h__
#define __sfm_h__

#ifdef __cplusplus
extern "C" {
#endif

#include "vector.h"

#define MAX_FLAT_REFRACTIONS 2
#define NUM_CAMERA_PARAMS (12+MAX_FLAT_REFRACTIONS)
//#define NUM_CAMERA_PARAMS 9
#define POLY_INVERSE_DEGREE 6

#define REFRACTION_TYPE_NONE 0 // no refraction
#define REFRACTION_TYPE_NONSHARED 1 // independent refraction interface for each camera
#define REFRACTION_TYPE_SHARED 2 // shared refraction interface for all cameras

typedef struct {
    double R[9];     /* Rotation */
    double t[3];     /* Translation */
    double f;        /* Focal length */
    double k[2];     /* Undistortion parameters */
    double k_inv[POLY_INVERSE_DEGREE]; /* Inverse undistortion parameters */
    char constrained[NUM_CAMERA_PARAMS];
    double constraints[NUM_CAMERA_PARAMS];  /* Constraints (if used) */
    double weights[NUM_CAMERA_PARAMS];      /* Weights on the constraints */
    double K_known[9];  /* Intrinsics (if known) */
    double k_known[5];  /* Distortion params (if known) */

    char fisheye;            /* Is this a fisheye image? */
    char known_intrinsics;   /* Are the intrinsics known? */
    double f_cx, f_cy;       /* Fisheye center */
    double f_rad, f_angle;   /* Other fisheye parameters */
    double f_focal;          /* Fisheye focal length */

    double f_scale, k_scale; /* Scale on focal length, distortion params */
	double d_scale;
	double n_scale;

	char num_refractions; /* Number of flat refraction planes in front of camera*/
	double n[3]; /* Normal of flat refraction planes, facing camera */
	double d[MAX_FLAT_REFRACTIONS]; /* Distance between each refraction plane */
	double mu[MAX_FLAT_REFRACTIONS+1]; /* Refractive index of each layer */
	char shared_interface_ref; /* Is this a reference camera for computing shared interface?*/
} camera_params_t;

/*
 * I wish closures were supported in C. This is the penalty for using
 * a callback function interface for the fzero optimization function.
 */
typedef struct {
	int num_refrac;
	double sx, sy;
	double *d, *mu;
} refract_back_proj_data;

/* Compute an updated rotation matrix given the initial rotation (R)
 * and the correction (w) */
void rot_update(double *R, double *w, double *Rnew);
    
void sfm_project(camera_params_t *init, double *K, 
		 double *w, double *dt, double *b, double *p,
		 int explicit_camera_centers);

void sfm_project_rd(camera_params_t *init, double *K, double *k,
                    double *R, double *dt, double *b, double *p,
                    int undistort, int explicit_camera_centers);

void sfm_project_refract(camera_params_t *init, double *K, double *k,
						 double *R, double *dt, double *b, double *p,
						 double *Rn, double *d, int undistort, 
						 int explicit_camera_centers);

void sfm_apply_rdist(camera_params_t *init, double *K, double *k, 
					 double *p);

/*
 * sz should be NEGATIVE so that it's in front of the camera (RHS coord system).
 * It's also measured from the first refractive interface rather than from the
 * camera.
 * d[0] is the distance from the camera to the first refractive interface.
 * The refractive interfaces are assumed to be parallel to the X-Y plane, and
 * the camera is at the origin. If this is not the case, you must first transform
 * the coordinate system.
 */
void sfm_solve_refrax(int num_refract, double sx, double sy, double sz, 
					  double *d, double *mu, double *x, double *y);

double sfm_solve_refrax_2d(double sx, double sy, double e, double r);

/* Solve refrax in 2D, multiple refractions, interfaces parallel to y axis.
 * sx is object distance along x axis, assumed positive and grater than sum(d)
 * Result is the refrax on the interface closest to the camera.
 */
double sfm_solve_multi_refrax_2d(int num_refrac, double sx, double sy, double *d, double *mu);

void sfm_update_Rn(double *n2, double *Rn);

v2_t sfm_project_final(camera_params_t *params, v3_t pt,
		       int explicit_camera_centers, int undistort);

v2_t sfm_project_refract_final(camera_params_t *cam, v3_t pt, 
							   int explicit_camera_centers, int undistort);

double refract_dist_limit(double *R, double *dt, double *n, int num_pts, 
		        	      double *B, int explicit_camera_centers);

/*
 * Compute orientation of the shared refraction interface, then rotate 
 * and translate the world coordinate system so that the refraction 
 * interface closest to the cameras is the XY plane.
 */
void sfm_transform_shared_interface(int num_cameras, int num_pts,
									camera_params_t *cams, v3_t *init_pts,
									v3_t *pt_constraints, 
									int use_point_constraints, 
									int explicit_camera_centers);

/*
 * refraction_type: see REFRACTION_TYPE definitions
 */
void run_sfm(int num_pts, int num_cameras, int ncons,
             char *vmask,
             double *projections,
             int est_focal_length,
			 int const_focal_length,
			 int refraction_type,
			 int est_refract_axis,
			 unsigned int refract_layers_mask,
			 int est_cam_t,
			 int est_cam_r,
			 int decouple_r_from_refract_axis,
             int undistort,
             int explicit_camera_centers,
             camera_params_t *init_camera_params,
             v3_t *init_pts, 
             int use_constraints, 
             int use_point_constraints,
             v3_t *points_constraints,
             double point_constraint_weight,
             int fix_points,
             int optimize_for_fisheye, 
             double eps2,
             double *Vout,
             double *Sout,
             double *Uout, double *Wout);

/* Refine the position of a single camera */
void camera_refine(int num_points, v3_t *points, v2_t *projs, 
		   camera_params_t *params, int adjust_focal,
                   int estimate_distortion);

/* Refine the position of a single camera, keeping the first fixed */
void camera_refine_fix_free(int num_points, v3_t *points, 
                            v2_t *projs0, v2_t *projs1,
                            camera_params_t *params0, 
                            camera_params_t *params1, double *H);

/*
 * Back-project ray in 2D. Refractive interfaces are parallel to the Y axis.
 *	d distance between each adjacent refractive interface. d[0] is distance 
 *		from camera to first interface.
 *	mu refractive index of each refractive layer
 *	sx object depth (positive, greater than sum(d))
 *	y y-coordinate of point to back project, at a distance 1 away from 
 *		the camera.
 */
double sfm_backProjectObj2D(int num_refrac, double *d, double *mu, double sx, double y);

/*
 * **** Everything is in camera coordinates!
 * p_n is ray direction
 * vout are refracted ray directions, including the original ray vout[0:2]
 *		Array size is (cam->num_refractions + 1) * 3
 * pout are ray origin points on the refractive interface
 *		Array size is cam->num_refractions * 3
 */
void sfm_refractiveBackProject(camera_params_t *cam, double *p_n, double *vout, double *pout);

/* 
 * Adapted from MATLAB
 * Return codes:
 *	1  Solution found
 *  -6 No sign change
 */
int sfm_fzero(double (*fun)(double, void*), double a, double b, double tolerance, void *data, double *ans);

#ifdef __cplusplus
}
#endif

#endif /* __sfm_h__ */
