%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%
% This file is adapted and modified from code accompanying the paper:
%   A. Agrawal, S. Ramalingam, Y. Taguchi, and V. Chari. A theory of 
%   multilayer flat refractive geometry. CVPR 2012.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% estimate tz and depth for a two layers given known mu, axis, R and vector s
% case 3 mu2 ~= mu0

% c is the condition number of the linear system being solved.

% d1 is optional and indicates known thickness of the second layer (between
% the first and second refractions). If not known, set to [].
function [d0,d1,t,c] = EstimateTzTwoLayeCase3KnownMu(n,R,s,mu1,mu2,xy,XYZ,idx,d1,xyb,mu1b,mu2b,d0)

if nargin < 10
    xyb = [];
    mu1b = [];
    mu2b = [];
end
    
N = size(XYZ,2);

% find rotation to make axis along -z axis
w = cross(n,[0;0;-1]);
if norm(w) ~= 0
    w = w/norm(w);
    theta = acos(n'*[0;0;-1]);
    w = theta*w;
    Rc = w2R(w);
else
    Rc = ones(3);
end
clear w

%check Rc*n should be [0;0;-1]
% Rc
% Rc * n
clear n



% rotate s = cross(n,t)
ss = Rc * s;
% now tx and tz can be obtained. We need to find tz
t = [-ss(2) ; ss(1) ; 0];
clear ss


% rotate 3D points
XYZ = Rc*R*XYZ + t*ones(1,N);

n2D = [0;-1];

% find z1, z2
z1 = [0;0;1];

%
zp = [0;1];

% TEST TEST Look at condition number versus number of points used
%idx = idx(1:2)

if(isempty(idx))
    idx = [1:N]';
end

Nsmall = size(idx,1);
if isempty(xyb)
    Mu1 = repmat(mu1,Nsmall,1);
    Mu2 = repmat(mu2,Nsmall,1);
    xy = xy(:,idx);
    XYZ = XYZ(:,idx);
else
    Mu1 = [repmat(mu1,Nsmall,1); repmat(mu1b,Nsmall,1)];
    Mu2 = [repmat(mu2,Nsmall,1); repmat(mu2b,Nsmall,1)];
    xy = [xy(:,idx), xyb(:,idx)];
    XYZ = [XYZ(:,idx), XYZ(:,idx)];
    Nsmall = Nsmall * 2;
end

if isempty(d1)
    A = zeros(Nsmall,3);
elseif isempty(d0)
    A = zeros(Nsmall,2);
else
    A = zeros(Nsmall,1);
end
b = zeros(Nsmall,1);

for jj = 1:Nsmall
    
    %ii = idx(jj,1);
    
    % camera ray
    %v = Rc*xy(:,ii);
    v = Rc*xy(:,jj);
    
    %P = XYZ(:,ii);
    P = XYZ(:,jj);
    
    %find z2
    z2 = cross(z1,cross(z1,v));
    z2 = z2/norm(z2);
    % make along x axis
    if(z2(1)<0)
        z2 = -z2;
        %disp('make along x axis');
    end
    
    %find projection
    vp = [z2'*v ; z1' * v];
    vp = vp/norm(vp);
    
    %find refracted ray
    v1 = RefractedRay(vp,n2D,1,Mu1(jj));
    
    % final refracted ray
    v2 = RefractedRay(v1,n2D,Mu1(jj),Mu2(jj));
    
    u = [z2'*P ; z1'*P];
    
    %make linear system
    if isempty(d1)
        b(jj,1) = -Cross2(v2,u);
        A(jj,:) = [-Cross2(v2,vp)/(vp'*zp) -Cross2(v2,v1)/(v1'*zp) Cross2(v2,zp)];
    elseif isempty(d0)
        b(jj,1) = -Cross2(v2,u) + d1*Cross2(v2,v1)/(v1'*zp);
        A(jj,:) = [-Cross2(v2,vp)/(vp'*zp) Cross2(v2,zp)];
    else
        b(jj,1) = -Cross2(v2,u) + d1*Cross2(v2,v1)/(v1'*zp) + d0*Cross2(v2,vp)/(vp'*zp);
        A(jj,:) = Cross2(v2,zp);
    end
end


c = cond(A);
x = A\b;

if isempty(d1)
    d0 = x(1);
    d1 = x(2);
    tz = x(3);
elseif isempty(d0)
    d0 = x(1);
    tz = x(2);
else
    tz = x(1);
end


% put tz in t
t(3) = tz;

% rotate back into original coordinate system
t = Rc'*t;













