%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%
% This file is adapted and modified from code accompanying the paper:
%   A. Agrawal, S. Ramalingam, Y. Taguchi, and V. Chari. A theory of 
%   multilayer flat refractive geometry. CVPR 2012.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Estimating refraction layer thicknesses assuming known refraction axis

function [bestsol, SolStats] = TzEstimationPlanarSceneKnownAxisCase3(rayDnoise,XYZ,mu1,mu2,n,d1known,rayDnoiseb,mu1b,mu2b,d0known)

if nargin < 7
    rayDnoiseb = [];
    mu1b = [];
    mu2b = [];
end

if nargin < 6
    d1known = [];
end

if nargin < 10
    d0known = [];
end

N = size(XYZ,2);

BestError = Inf;
val = Inf;

if(~exist('RANSAC_trials','var'))
    RANSAC_trials = 300;
end


RANSAC_trials

Rz = vrrotvec2mat(vrrotvec(-n, [0; 0; 1]));
rays = Rz * rayDnoise;

if isempty(rayDnoiseb)
    % Theoretically, 5 points is the minimum required by planeRefracPose
    numEstPoints = 20;
    raysb = []
else
    numEstPoints = 20;
    raysb = Rz * rayDnoiseb;
end

% One row for each estimate produced. Columns are:
% trial num, trial solution num, 3d error (-1 if not est), condition num,
% d0, d1, tx, ty, tz
SolStats = zeros(0,9);

for trial = 1:RANSAC_trials
    idx = unique(randi(N,numEstPoints,1));
    while(size(idx,1) < numEstPoints)
        idx = unique(randi(N,numEstPoints,1));
    end
    
    if isempty(rayDnoiseb)
        [R,t,AllR,Allt] = planeRefracPose(XYZ(:,idx)', rays(:,idx)');
    else
        [R,t,AllR,Allt] = planeRefracPose(repmat(XYZ(:,idx)',2,1), [rays(:,idx)'; raysb(:,idx)']);
    end
    
    numSoln = size(AllR,3);
    
    for solNum = 1:numSoln
        R = Rz' * AllR(:,:,solNum);
        t = Rz' * [Allt(:,:,solNum); 0];

        [d0,d1,t,condNum] = EstimateTzTwoLayeCase3KnownMu(n,R,cross(n,t),mu1,mu2,rayDnoise,XYZ,idx,d1known,rayDnoiseb,mu1b,mu2b,d0known);

        tz = t(3,:)';

        SolStats = [SolStats; trial, solNum, -1, condNum, d0, d1, t'];
        
        % d should be greater than 0 and tz, translation in z direction should
        % be greater than d
        solIdx = find(d0 > 0 & d1 > 0 & tz > (d0+d1));
        if(isempty(solIdx))
%             disp(['solNum=' num2str(solNum) ': no solution found for d and tx']);
            continue;
        end

        Rsol = R;

        % find solution with minimum error using all points
        val = Compute3DErrorTwoLayerCase3(t,rayDnoise,XYZ,Rsol,n,mu1,mu2,d0,d1,rayDnoiseb,mu1b,mu2b);    

        SolStats(end,3) = val;
        
        solVec = [n ; d0 ; Rot2RPY(R) ; t ; d1];

        if(val < BestError)
            BestError = val;
            bestsol = solVec;
%             fprintf('3D MSE Error 8 point planar scene  = %f\n',val);
%             d0

            if(val < 1e-6)
                break;
            end
        end
    end
    
    if(val < 1e-6)
        break;
    end
end

