%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [] = alignGridPoints(displayFigures)

    SetDefaultValue(1, 'displayFigures', 0);

    ptsDataFile = 'GridPointData.mat';
    saveDataFile = 'AlignedGridPointData.mat';
    
    xlimits = [0 1024];
    ylimits = [0 768];
    
    % Run detectGridPts2.m or detectCheckerPoints.m to generate this data
    load(ptsDataFile, 'PointDataR', 'PointDataB', 'allImages', 'camIDs', 'seqNums', 'fileExt', 'sep');
    
    if length(camIDs) > 1
        error(['File ' pointsData ' contains data for more than one camera.']);
    end
    
    % Find grid assignment for red and blue points such that the two grids
    % are matched up
    AssignedGridR = cell(0,1);
    AssignedGridB = cell(0,1);
    AssignedPointsR = cell(0,1);
    AssignedPointsB = cell(0,1);
    AssignedPointFiles = cell(0,1);
    matchDist = 13; % Distance for matching coordinate origin point    
    
    for i=1:numel(camIDs)
        for j=1:numel(seqNums)
            imgFile = [seqNums{j} sep camIDs{i} '.' fileExt];
    
            if ~exist(imgFile, 'file')
                break;
            end
            
            disp(['Found file ' imgFile]);
    
            nPtsB = size(PointDataB{i,j},1);
            nPtsR = size(PointDataR{i,j},1);
        
            if (nPtsB == 0) || (nPtsR == 0)
                disp('  Insufficient detected points for this file.');
                break;
            end
            
            % Detected points are in 1-based indexing, but the calibration
            % toolbox uses 0-based indexing.
            PointsB = PointDataB{i,j} - ones(nPtsB,2);
            PointsR = PointDataR{i,j} - ones(nPtsR,2);
        
            GridR = assignGridPoints2(PointsR);
            GridB = assignGridPoints2(PointsB);
        
            if any(any(isnan(GridR)))
                disp(['Warning: point assignment failed for red grid in file ' imgFile]);
            end
            
            if any(any(isnan(GridB)))
                disp(['Warning: point assignment failed for blue grid in file ' imgFile]);
            end
            
            ciB = find((GridB(:,1) == 0) & (GridB(:,2) == 0));
            [xiB, xDist] = nextGridPoint(0, GridB);
            [yiB, yDist] = nextGridPoint(1, GridB);

            cB = PointsB(ciB,:);
            xB = (PointsB(xiB,:) - cB) / xDist;
            yB = (PointsB(yiB,:) - cB) / yDist;

            ciR = find((GridR(:,1) == 0) & (GridR(:,2) == 0));
            [xiR, xDist] = nextGridPoint(0, GridR);
            [yiR, yDist] = nextGridPoint(1, GridR);

            cR = PointsR(ciR,:);
            xR = (PointsR(xiR,:) - cR) / xDist;
            yR = (PointsR(yiR,:) - cR) / yDist;

            if abs(xB * xR') < abs(xB * yR')
                % The blue x-axis is actually parallel to the red y-axis. So
                % flip x and y for red.
                tmp = GridR(:,1);
                GridR(:,1) = GridR(:,2);
                GridR(:,2) = tmp;

                tmp = xiR;
                xiR = yiR;
                yiR = tmp;

                tmp = xR;
                xR = yR;
                yR = tmp;

%                 disp('Swapping x and y');
            end

            if (xB * xR') < 0
                % Red x-axis is pointing in the opposite direction.
                GridR(:,1) = GridR(:,1) * -1;
%                 disp('Flipping x axis');
            end

            if (yB * yR') < 0
                GridR(:,2) = GridR(:,2) * -1;
%                 disp('Flipping y axis');
            end

            ciBMatch = find((abs(PointsR(:,1) - cB(1)) <= matchDist) & ...
                (abs(PointsR(:,2) - cB(2)) <= matchDist));

            if length(ciBMatch) ~= 1
                if length(ciBMatch) == 0
                    error('No match for grid coordinate origin');
                else
                    error('Too many matches for grid coordinate origin');
                end
            else
                gridOffset = GridR(ciBMatch,:);
                GridR = GridR - ones(nPtsR,1) * gridOffset;

                if displayFigures
                    figure, hold on, axis equal;
                    for j=1:nPtsR
                        text(PointsR(j,1), PointsR(j,2)+6, [' ' num2str(GridR(j,1)) ', ' num2str(GridR(j,2))], 'FontSize', 6);
                    end
                    plot(PointsR(:,1), PointsR(:,2), '*');
                    xlim(xlimits), ylim(ylimits);
                    hold off;
                    title(['Red points for file ' imgFile], 'Interpreter', 'none');

                    figure, hold on, axis equal;
                    for j=1:nPtsB
                        text(PointsB(j,1), PointsB(j,2)+6, [' ' num2str(GridB(j,1)) ', ' num2str(GridB(j,2))], 'FontSize', 6);
                    end
                    plot(PointsB(:,1), PointsB(:,2), '*');
                    xlim(xlimits), ylim(ylimits);
                    hold off;
                    title(['Blue points for file ' imgFile], 'Interpreter', 'none');
                end
            end

            AssignedGridR{end+1} = GridR;
            AssignedGridB{end+1} = GridB;
            
            AssignedPointsR{end+1} = PointsR;
            AssignedPointsB{end+1} = PointsB;
            
            AssignedPointFiles{end+1} = imgFile;
        end
    end
    
    save(saveDataFile, 'AssignedGridR', 'AssignedPointsR', 'AssignedGridB', 'AssignedPointsB', 'AssignedPointFiles', 'matchDist');
end

    
function [idx, dist] = nextGridPoint(dir, assignedGrid)
    if dir == 0
        candidates = ((assignedGrid(:,1) > 0) & (assignedGrid(:,2) == 0));
    else
        candidates = ((assignedGrid(:,1) == 0) & (assignedGrid(:,2) > 0));
    end
    
    if ~any(candidates)
        error(['Could not find next grid point along direction ' num2str(dir)]);
    end
    
    cIdx = find(candidates);
    dists = sum(assignedGrid(candidates,:), 2);
    [unused, sIdx] = sort(dists);
    
    idx = cIdx(sIdx(1));
    dist = dists(sIdx(1));
end