%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [gridCoord] = assignGridPoints2(X)
% Given a set of 2D image points, try to deduce a rectangular grid and
% assign the points to an axis-aligned grid model.
%   X is N-by-2
%   S is N-by-2

    n = size(X,1);

    % Choose origin as the point closest to the middle
    centroid = sum(X,1)/n;
    [notused, originIdx] = min(sum((X - ones(n,1)*centroid).^2,2));
    originPt = X(originIdx,:);
    X = X - ones(n,1) * originPt;
    originPt = [0, 0];
    
    % Pick two closest points that are separated by between 60 and 120
    % degrees. These will define the grid axes.
    originDists = sqrt(sum(X.^2,2));
    [B,IX] = sort(originDists);
    n1 = X(IX(2),:);
    
    n2 = nan;
    for i=3:5
        tmp = X(IX(i),:);
        angle = acos((tmp * n1') / (norm(tmp)*norm(n1)));
        if (angle > (60*pi/180)) && (angle < (120*pi/180))
            n2 = tmp;
        end
    end
    
    if isnan(n2)
        error('Could not find grid axes');
    end
    
    % Make n1 the x axis and n2 the y axis
    if abs(n1 * [1;0]) < abs(n2 * [1;0])
        tmp = n1;
        n1 = n2;
        n2 = tmp;
    end
    
    if (n1 * [1;0]) < 0
        n1 = n1 * -1;
    end
    
    if (n2 * [0;1]) < 0
        n2 = n2 * -1;
    end
    
    margin = 10;
    xlimits = [min(X(:,1))-margin, max(X(:,1))+margin];
    ylimits = [min(X(:,2))-margin, max(X(:,2))+margin];

    gridCoord = nan(n,2);
    [coord, idx] = assignLineCoords(X, originPt-n1, n1, xlimits, ylimits);
    gridCoord(idx,1) = coord - 1;
    gridCoord(idx,2) = 0;
    
    [coord, idx2] = assignLineCoords(X, originPt, -n1, xlimits, ylimits);
    gridCoord(idx2,1) = -coord;
    gridCoord(idx2,2) = 0;
    
    idx = [idx; idx2];
    gridCoord = assignMultiLineCoords(X, gridCoord, idx, n2, 1, xlimits, ylimits);
    
    idx = find(gridCoord(:,1) == 0);
    gridCoord = assignMultiLineCoords(X, gridCoord, idx, n1, 0, xlimits, ylimits);
    
%     figure, hold on, axis equal;
%     plot(X(:,1), X(:,2), '*');
%     for i=1:n
%         if ~isnan(gridCoord(i,1))
%             text(X(i,1), X(i,2)+6, [' ' num2str(gridCoord(i,1)) ', ' num2str(gridCoord(i,2))], 'FontSize', 6);
%         end
%     end
%     plot([0, n1(1)], [0, n1(2)], 'r');
%     plot([0, n2(1)], [0, n2(2)], 'r');
%     hold off;
    
    return;
end

function [coords] = assignMultiLineCoords(X, initCoords, startIdx, dir, dirWhichAxis, xlimits, ylimits)
    % startCoords is N-by-2 where N is the total number of points in X
    
    coords = initCoords;
    
    for i=1:length(startIdx)
        [lineCoords, newIdx] = assignLineCoords(X, X(startIdx(i),:), dir, xlimits, ylimits);
        [lineCoords2, newIdx2] = assignLineCoords(X, X(startIdx(i),:), -dir, xlimits, ylimits);
        
        lineCoords = [lineCoords; -lineCoords2];
        newIdx = [newIdx; newIdx2];
        
        numFound = length(lineCoords);
        
        if dirWhichAxis == 0
            newCoords = [lineCoords, initCoords(startIdx(i),2) * ones(numFound,1)];
        else
            newCoords = [initCoords(startIdx(i),1) * ones(numFound,1), lineCoords];
        end
        
        if any(any((coords(newIdx,:) ~= newCoords) & ~isnan(coords(newIdx,:))))
            error('Conflicting coordinates being assigned');
        end
        
        coords(newIdx,:) = newCoords;
    end
end
    
function [lineCoords, idx] = assignLineCoords(X, startPt, dir, xlimits, ylimits)
    n = size(X,1);
    tol = norm(dir) / 2;
    extrapolateLimit = 2; % Max number of steps with no hits
    stepsSinceHit = 0;
    totalSteps = 1;
    
    lineCoords = zeros(0,1);
    idx = zeros(0,1);
    
    while true
        probePt = startPt + dir;
        
        if probePt(1) > xlimits(2) || probePt(1) < xlimits(1) || ...
                probePt(2) > ylimits(2) || probePt(2) < ylimits(1)
            break;
        end
        
        distSq = sum((X - ones(n,1) * probePt).^2,2);
        [minDist, ix] = min(distSq);
        nextPt = X(ix,:);
        
        if minDist > tol
            stepsSinceHit = stepsSinceHit + 1;
            startPt = probePt;
            
            if stepsSinceHit > extrapolateLimit
                break;
            end
        else
            stepsSinceHit = 0;
            
            lineCoords(end+1,1) = totalSteps;
            idx(end+1,1) = ix;
            
            dir = nextPt - startPt;
            startPt = nextPt;
        end
        
        totalSteps = totalSteps + 1;
    end
end
