%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [R,t,AllR,Allt] = planeRefracPose(S, rays)
% S is an N-by-2 matrix of 2D model points, N >= 5
% rays is N-by-3 matrix of the corresponding 3D ray directions
%
% R and t are the rotation and translation of the model points. 
% 
% The refraction interface(s) are assumed to be perpendicular to the Z
% axis. The camera is looking down the positive Z axis.
%
% There are up to four ambiguous solutions, two in front of the camera 
% and two behind it. This ambiguity can be resolved by estimating the 
% refraction layer thicknesses and the z-position of the object.

    epsilon = 1e-12;

    rays = rays ./ (sqrt(sum(rays.^2,2))*ones(1,3));
    
    B = [S(:,1).*rays(:,2), -S(:,1).*rays(:,1), S(:,2).*rays(:,2), -S(:,2).*rays(:,1), rays(:,2), -rays(:,1)];

    [U,Z,V] = svd(B);
    x = V(:,end);

    %     [k*a, k*b, +/-sqrt(1-k^2*a^2-k^2*b^2)
    % R =  k*c, k*d, +/-sqrt(1-k^2*c^2-k^2*d^2)
    %      ?  , ?  , ?                         ]
    %
    % Third row of R is recovered as the cross product of the first two
    % rows.
        
    a = x(1); c = x(2); b = x(3); d = x(4);
    
    p2 = 2*a*b*c*d-a*a*d*d-b*b*c*c;
    p1 = a^2+b^2+c^2+d^2;
    p0 = -1;
    
    disc = p1^2 - 4*p2*p0;
    factors = sqrt(roots([p2, p1, p0]));

    if (imag(factors(1)) ~= 0) && (abs(disc) < epsilon)
        % Discriminant is essentially zero, so the quadratic probably has
        % a unique root.
        factors = real(factors(1));
    end

    pm = 0; % plus/minus for R(1,3)*R(2,3)
    toler = 1e-6;
    foundSoln = 0;
    
    for i=1:length(factors) % only up to 2
        k = factors(i);

        % Form the two rows of R from a,b,c,d and scale factor. Check that
        % they are unit length and orthogonal.
        z1 = k^2*a^2 + k^2*b^2;
        z2 = k^2*c^2 + k^2*d^2;
        if (z1 <= 1+epsilon) && (z2 <= 1+epsilon)
            if z1 > 1
                z1 = 0;
            else
                z1 = sqrt(1 - z1);
            end
            
            if z2 > 1
                z2 = 0;
            else
                z2 = sqrt(1 - z2);
            end

            for pm = -1:2:1
                if abs(k^2*a*c + k^2*b*d + pm*z1*z2) < toler
                    foundSoln = true;
                    break;
                end
            end
        end
    end

    if ~foundSoln
        a
        b
        c
        d
        p2
        p1
        p0
        disc
        factors
        error('No solution found within tolerance!');
    end

    % One possible solution
    R = makeR(a, b, c, d, k, z1, pm*z2);    
    t = x(5:6) * k;
    
    % Multiple ambiguous solutions
    AllR = zeros(3,3,4);
    Allt = zeros(2,1,4);
    
    AllR(:,:,1) = R;
    Allt(:,:,1) = t;
    
    AllR(:,:,2) = makeR(a, b, c, d, -k, z1, pm*z2);
    Allt(:,:,2) = -t;

    if pm == 1
        pm1 = -1; pm2 = -1;
    else
        pm1 = -1; pm2 = 1;
    end
    
    AllR(:,:,3) = makeR(a, b, c, d, k, pm1*z1, pm2*z2);
    Allt(:,:,3) = t;
        
    AllR(:,:,4) = makeR(a, b, c, d, -k, pm1*z1, pm2*z2);
    Allt(:,:,4) = -t;
    
       
function [R] = makeR(a, b, c, d, k, z1, z2)
    R = zeros(3);
    R(1) = a*k;
    R(2) = c*k;
    R(4) = b*k;
    R(5) = d*k;
    R(7) = z1;
    R(8) = z2;
    R(3,:) = cross(R(1,:), R(2,:));

    % Find closest orthogonal matrix
    [U,Z,V] = svd(R);
    R = U * V';
    