%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%
% The parameterization of the nonlinear optimization is based on code
% that accompanies the paper:
%   A. Agrawal, S. Ramalingam, Y. Taguchi, and V. Chari. A theory of 
%   multilayer flat refractive geometry. CVPR 2012.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all;close all;clc;format long;

rawPointsData = 'GridPointData.mat';
pointsData = 'AlignedGridPointData.mat';

outFile = 'CalibResult.mat';

load(rawPointsData, 'camIDs');

if numel(camIDs) > 1
    error(['File ' pointsData ' contains data for more than one camera. Refractive calib is for one camera at a time.']);
end

lcaFolder = ['../../lca/' camIDs{1}];
calibRed = [lcaFolder '/result/Calib_Results_red.mat'];
calibBlue = [lcaFolder '/result/Calib_Results_blue.mat'];
lcaColorAlignData = [lcaFolder '/ColorAlignment.mat'];

% Image width and height
load(calibRed, 'nx', 'ny');
xlimits = [0 nx]; 
ylimits = [0 ny];
clear nx ny;

% Water 19deg.
muWaterRed = 1.33151;
muWaterBlue = 1.34318;
% Water 21.5deg.
%muWaterRed = 1.33129;
%muWaterBlue = 1.34295;
% Water 24deg.
%muWaterRed = 1.33104;
%muWaterBlue = 1.34269;
muGlassRed = 1.488;
muGlassBlue = 1.516; 
glassThickness = 0.0056; % in metres
gridSize = 0.00599; % in metres

load(pointsData, 'AssignedPointsR', 'AssignedPointsB', 'AssignedGridR', 'AssignedGridB', 'AssignedPointFiles');
numImgs = numel(AssignedPointsR);

load(calibBlue, 'KK', 'fc', 'kc', 'cc', 'alpha_c');
KKb = KK;
fcb = fc;
kcb = kc;
ccb = cc;
alpha_cb = alpha_c;
    
load(calibRed, 'KK', 'fc', 'kc', 'cc', 'alpha_c');
KKr = KK;
fcr = fc;
kcr = kc;
ccr = cc;
alpha_cr = alpha_c;

clear KK fc kc cc alpha_c;

disp('Undistorting points...');

load(lcaColorAlignData, 'blueScale', 'blueShift', 'blueKcWarp');

% Pre-process data to remove distortion and decentering, and keep only
% points common to both R and B sets
for i = 1:numImgs
    [C, ia, ib] = intersect(AssignedGridR{i}, AssignedGridB{i}, 'rows');
    AssignedPointsR{i} = AssignedPointsR{i}(ia,:);
    AssignedPointsB{i} = AssignedPointsB{i}(ib,:);
    AssignedGridR{i} = C;
    
    % Updates to the new KKr/KKb as well
    [AssignedPointsR{i}, KKr] = undistortAndCenterPoints(AssignedPointsR{i}, fcr, ccr, kcr, alpha_cr, xlimits(2), ylimits(2));
    [AssignedPointsB{i}, KKb] = undistortAndCenterPoints(AssignedPointsB{i}, fcb, ccb, kcb, alpha_cb, xlimits(2), ylimits(2));
    
    AssignedPointsB{i} = undistortAndScale(AssignedPointsB{i}, blueScale, blueShift, [KKb(1,1); KKb(2,2)], [KKb(1,3); KKb(2,3)], blueKcWarp, 0);
end

disp('done.');

clear fcr ccr kcr alpha_cr fcb ccb kcb alpha_cb;

XInit = zeros(numImgs, 9+3);
XFinal = zeros(numImgs, 9+3);

Reproj = zeros(numImgs, 1);
ReprojR = zeros(numImgs, 1);
ReprojB = zeros(numImgs, 1);
ReprojFinal = zeros(numImgs, 1);
ReprojFinalR = zeros(numImgs, 1);
ReprojFinalB = zeros(numImgs, 1);

PtsReprojB = cell(numImgs,1);
PtsReprojR = cell(numImgs,1);
PtsReprojFinalR = cell(numImgs,1);
PtsReprojFinalB = cell(numImgs,1);

Ptsr = cell(numImgs,1);
Ptsb = cell(numImgs,1);

for i=1:numImgs
    seqNum = strtok(AssignedPointFiles{i}, '_');
    
    Xr = AssignedPointsR{i};
    Xb = AssignedPointsB{i};
    
    % For saving afterwards
    Ptsr{i} = Xr;
    Ptsb{i} = Xb;
    
    % Assigned grid for red and blue are the same
    N = size(AssignedGridR{i},1);
    Pall = [AssignedGridR{i}' * gridSize; zeros(1,N)];
    
    CameraRays = [Xr,ones(N,1)]';
    CameraRaysb = [Xb,ones(N,1)]';
    
    ray2D = KKr\CameraRays;
    ray2Db = KKr\CameraRaysb;

    [A, axisPt] = axisFromDisp(Xr, Xb, KKr);
    A = -A;

    solVec = TzEstimationPlanarSceneKnownAxisCase3(ray2D,Pall,muGlassRed,muWaterRed,A,glassThickness,ray2Db,muGlassBlue,muWaterBlue);
    neInit = solVec(1:3);
    tau1Init = solVec(4);
    RInit = RPY2Rot(solVec(5:7));
    tInit = solVec(8:10);
    wInit = solVec(5:7);
    tau2Init = solVec(11);

    error3DInit = Compute3DErrorTwoLayerCase3(tInit,ray2D,Pall,RPY2Rot(wInit),neInit,muGlassRed,muWaterRed,tau1Init,tau2Init,ray2Db,muGlassBlue,muWaterBlue)

    initAxis = A(1:2) / abs(A(3));
    %non-linear optimization
    x0 = [initAxis ; wInit ; tInit ; tau1Init ; ];
    XInit(i,:) = [x0; neInit]';

    x0Scaled = x0;
        
%     [PtsReprojR{i}, PtsReprojB{i}] = PlotReprojectedPointsNoimgCase3Disp(x0,CameraRays,CameraRaysb,Pall,muGlassRed,muWaterRed,muGlassBlue,muWaterBlue,tau2Init,KKr,KKr);  
%     [Reproj(i), ReprojR(i), ReprojB(i)] = ReprojectionErrorTwoLayersCase3Disp(neInit,wInit,tInit,tau1Init,tau2Init,CameraRays,Pall,KKr,muGlassRed,muWaterRed,CameraRaysb,muGlassBlue,muWaterBlue,KKr);
%     str = sprintf('Initial Solution: Reprojection error = %f pixels\n',Reproj(i));
%     title(str);
%     xlim(xlimits);
%     ylim(ylimits);
%     axisPtInit = KKr * neInit;
%     axisPtInit = axisPtInit / axisPtInit(3)
%     plot(axisPtInit(1), axisPtInit(2), 'gx', 'MarkerSize', 10);
%     set(gca, 'YDir', 'reverse');
    
    options = optimset('Display','iter','TolFun',1e-6,'TolX',1e-6,'MaxIter',5000,'MaxFunEvals',10000);
    x = lsqnonlin(@CostFuncTwoLayersCase3Disp,x0Scaled,[],[],options,CameraRays,CameraRaysb,Pall,KKr,muGlassRed,muWaterRed,muGlassBlue,muWaterBlue,tau2Init,KKr);

    neFinal = [x(1:2) ; -1];
    neFinal = neFinal/norm(neFinal);
    wFinal = x(3:5);
    tFinal = x(6:8);
    tau1Final = x(9);
    tau2Final = tau2Init; % Not estimated, assumed known
    
    XFinal(i,:) = [x; neFinal];

    error3DFinal = Compute3DErrorTwoLayerCase3(tFinal,ray2D,Pall,RPY2Rot(wFinal),neFinal,muGlassRed,muWaterRed,tau1Final,tau2Final,ray2Db,muGlassBlue,muWaterBlue)

%     [PtsReprojFinalR{i}, PtsReprojFinalB{i}] = PlotReprojectedPointsNoimgCase3Disp(x,CameraRays,CameraRaysb,Pall,muGlassRed,muWaterRed,muGlassBlue,muWaterBlue,tau2Final,KKr,KKr);
%     [ReprojFinal(i), ReprojFinalR(i), ReprojFinalB(i)] = ReprojectionErrorTwoLayersCase3Disp(neFinal,wFinal,tFinal,tau1Final,tau2Final,CameraRays,Pall,KKr,muGlassRed,muWaterRed,CameraRaysb,muGlassBlue,muWaterBlue,KKr);
%     str = sprintf('Final Solution after non-linear refinement: Reprojection error = %f pixels\n',ReprojFinal(i));
%     title(str);
%     xlim(xlimits);
%     ylim(ylimits);
%     axisPtFinal = KKr * neFinal;
%     axisPtFinal = axisPtFinal / axisPtFinal(3);
%     plot(axisPtFinal(1), axisPtFinal(2), 'gx', 'MarkerSize', 10);
%     set(gca, 'YDir', 'reverse');
end

save(outFile, 'XInit', 'XFinal', 'Reproj', 'ReprojR', 'ReprojB', 'ReprojFinal', 'ReprojFinalR', 'ReprojFinalB', 'PtsReprojR', 'PtsReprojB', 'PtsReprojFinalR', 'PtsReprojFinalB', 'Ptsr', 'Ptsb', 'muWaterRed', 'muWaterBlue', 'muGlassRed', 'muGlassBlue', 'glassThickness', 'gridSize', 'KKr', 'KKb');

disp('All done.');
return;

