%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Thesis title: Underwater Camera Calibration and 3D Reconstruction
% http://hdl.handle.net/10402/era.38695
% Author: Timothy Yau
%
% This file is based on code from the Camera Calibration Toolbox by 
% Jean-Yves Bouguet.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [Irec] = rectWithBundler(I,R,f,c,k,alpha,KK_new,nrNew,ncNew,fB,kB)
% fB and kB are the Bundler-computed focal length and radial distortion
% parameters. 

[nr,nc] = size(I);

[mx,my] = meshgrid(1:ncNew, 1:nrNew);
px = reshape(mx',ncNew*nrNew,1);
py = reshape(my',ncNew*nrNew,1);
clear mx my;

% First transform according to Bundler's radial distortion formula to get
% the sample locations in the undistorted image (according to separate
% intrinsic calibration), then apply the undistortion from calibration to
% get sample locations in the original image.
% See RadialUndistort.cpp for the formula.
px_c = px - 0.5 * ncNew; % Off by 0.5 ??
py_c = py - 0.5 * nrNew;
clear px py;

r2 = (px_c .* px_c + py_c .* py_c) / (fB * fB);
factor = 1.0 + kB(1) .* r2 + kB(2) .* r2 .* r2;
px_c = px_c .* factor;
py_c = py_c .* factor;
clear factor;
clear r2;

px_c = px_c + 0.5 * ncNew;
py_c = py_c + 0.5 * nrNew;

rays = KK_new\[(px_c - 1)';(py_c - 1)';ones(1,length(px_c))];
clear px_c py_c;


% Rotation: (or affine transformation):

rays2 = R'*rays;
clear rays;

x = [rays2(1,:)./rays2(3,:);rays2(2,:)./rays2(3,:)];
clear rays2;

% Add distortion:
xd = apply_distortion(x,k);
clear x;

% Reconvert in pixels:

px2 = f(1)*(xd(1,:)+alpha*xd(2,:))+c(1);
py2 = f(2)*xd(2,:)+c(2);
clear xd;

% Interpolate between the closest pixels:

px_0 = floor(px2);


py_0 = floor(py2);
%py_1 = py_0 + 1;


good_points = find((px_0 >= 0) & (px_0 <= (nc-2)) & (py_0 >= 0) & (py_0 <= (nr-2)));

px2 = px2(good_points);
py2 = py2(good_points);
px_0 = px_0(good_points);
py_0 = py_0(good_points);

alpha_x = px2 - px_0;
alpha_y = py2 - py_0;
clear px2 py2;

a1 = (1 - alpha_y).*(1 - alpha_x);
a2 = (1 - alpha_y).*alpha_x;
a3 = alpha_y .* (1 - alpha_x);
a4 = alpha_y .* alpha_x;
clear alpha_y alpha_x;

ind_lu = px_0 * nr + py_0 + 1;
ind_ru = (px_0 + 1) * nr + py_0 + 1;
ind_ld = px_0 * nr + (py_0 + 1) + 1;
ind_rd = (px_0 + 1) * nr + (py_0 + 1) + 1;
clear px_0 py_0;

[mx,my] = meshgrid(1:ncNew, 1:nrNew);
px = reshape(mx',ncNew*nrNew,1);
py = reshape(my',ncNew*nrNew,1);
clear mx my;

ind_new = (px(good_points)-1)*nrNew + py(good_points);
clear good_points px py;

Irec = 255*zeros(nrNew,ncNew);

Irec(ind_new) = a1 .* I(ind_lu) + a2 .* I(ind_ru) + a3 .* I(ind_ld) + a4 .* I(ind_rd);


return;

